# pyre-unsafe

class Manager:
    def aggregate(self, *args, **kwargs): ...
    def all(self): ...
    def create(self, **kwargs): ...
    def filter(self, *args, **kwargs): ...
    def get(self, *args, **kwargs): ...
    def get_or_create(self, defaults=None, **kwargs): ...
    def raw(self, raw_query, params=None, translations=None, using=None): ...
    def bulk_create(self, *args): ...
    def using(self, *args): ...
    def defer(self, *args, **kwargs): ...



from logging import Logger
from threading import Condition, Lock
from typing import Any, Deque, Mapping, Optional, Sequence, Set, Tuple

from .buffers import ReadOnlyFileBasedBuffer
from .channel import HTTPChannel
from .compat import reraise, tobytes
from .utilities import Error

rename_headers: Mapping[str, str]
hop_by_hop: frozenset

class ThreadedTaskDispatcher:
    stop_count: int = ...
    active_count: int = ...
    logger: Logger = ...
    queue_logger: Logger = ...
    threads: Set = ...
    queue: Deque[Task] = ...
    lock: Lock = ...
    queue_cv: Condition = ...
    thread_exit_cv: Condition = ...
    def __init__(self) -> None: ...
    def start_new_thread(self, target: Any, args: Any) -> None: ...
    def handler_thread(self, thread_no: int) -> None: ...
    def set_thread_count(self, count: int) -> None: ...
    def add_task(self, task: Task) -> None: ...
    def shutdown(self, cancel_pending: bool = ..., timeout: int = ...) -> bool: ...

class Task:
    close_on_finish: bool = ...
    status: str = ...
    wrote_header: bool = ...
    start_time: int = ...
    content_length: Optional[int] = ...
    content_bytes_written: int = ...
    logged_write_excess: bool = ...
    logged_write_no_body: bool = ...
    complete: bool = ...
    chunked_response: bool = ...
    logger: Logger = ...
    channel: HTTPChannel = ...
    request: Error = ...
    response_headers: Sequence[Tuple[str, str]] = ...
    version: str = ...
    def __init__(self, channel: HTTPChannel, request: Error) -> None: ...
    def service(self) -> None: ...
    @property
    def has_body(self) -> bool: ...
    def build_response_header(self) -> bytes: ...
    def remove_content_length_header(self) -> None: ...
    def start(self) -> None: ...
    def finish(self) -> None: ...
    def write(self, data: bytes) -> None: ...

class ErrorTask(Task):
    complete: bool = ...
    status: str = ...
    close_on_finish: bool = ...
    content_length: int = ...
    def execute(self) -> None: ...

class WSGITask(Task):
    environ: Optional[Any] = ...
    response_headers: Sequence[Tuple[str, str]] = ...
    complete: bool = ...
    status: str = ...
    content_length: int = ...
    close_on_finish: bool = ...
    def execute(self) -> None: ...
    def get_environment(self) -> Any: ...



# pyre-unsafe

DEFAULT_TIMEOUT = object()

class BaseCache:
    def __init__(self, params): ...
    def get_backend_timeout(self, timeout=DEFAULT_TIMEOUT): ...
    def make_key(self, key, version=None): ...
    def add(self, key, value, timeout=DEFAULT_TIMEOUT, version=None): ...
    def get(self, key, default=None, version=None): ...
    def set(self, key, value, timeout=DEFAULT_TIMEOUT, version=None): ...
    def delete(self, key, version=None): ...
    def get_many(self, keys, version=None): ...
    def has_key(self, key, version=None): ...
    def incr(self, key, delta=1, version=None): ...
    def decr(self, key, delta=1, version=None): ...
    def __contains__(self, key): ...
    def set_many(self, data, timeout=DEFAULT_TIMEOUT, version=None): ...
    def delete_many(self, keys, version=None): ...
    def clear(self): ...
    def validate_key(self, key): ...
    def incr_version(self, key, delta=1, version=None): ...
    def decr_version(self, key, delta=1, version=None): ...
    def close(self, **kwargs): ...



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import functools
import logging
import os
import shutil
import subprocess
import textwrap
from itertools import chain
from pathlib import Path
from time import time
from typing import ContextManager, Dict, List, NamedTuple, Optional, Set, Tuple

from . import buck, json_rpc, log
from .buck import BuckBuilder, find_buck_root
from .configuration import Configuration
from .exceptions import EnvironmentException
from .filesystem import (
    _compute_symbolic_link_mapping,
    _delete_symbolic_link,
    acquire_lock,
    acquire_lock_if_needed,
    add_symbolic_link,
    do_nothing,
    find_python_paths,
    is_empty,
    is_parent,
    remove_if_exists,
    translate_path,
    translate_paths,
)
from .find_directories import CONFIGURATION_FILE, LOCAL_CONFIGURATION_FILE
from .socket_connection import SocketConnection, SocketException


LOG: logging.Logger = logging.getLogger(__name__)


# If there are a lot of tracked files that are updated at the same time, it is
# probably a rebase. So, rebuild just to be safe.
REBUILD_THRESHOLD_FOR_UPDATED_PATHS: int = 50


# If there are a lot of new or deleted files, it is probably a rebase.
# This is a separate threshold from the number of updated tracked files because
# we don't know for sure if these new files should actually be tracked. So, we
# might want a higher threshold in order to avoid rebuilding for spurious new
# files.
REBUILD_THRESHOLD_FOR_NEW_OR_DELETED_PATHS: int = 50


DONT_CARE_PROGRESS_VALUE = 1

BUCK_BUILDER_CACHE_PREFIX = ".buck_builder_cache"


READER_WRITER_LOCK = "analysis_directory_reader_writer.lock"


class NotWithinLocalConfigurationException(Exception):
    pass


def _resolve_filter_paths(
    source_directories: List[str],
    targets: List[str],
    configuration: Configuration,
    original_directory: str,
) -> Set[str]:
    filter_paths = set()
    if source_directories or targets:
        if source_directories:
            filter_paths.update(source_directories)
        if targets:
            filter_paths.update(
                [buck.presumed_target_root(target) for target in targets]
            )
    else:
        local_configuration_root = configuration.local_root
        if local_configuration_root:
            filter_paths = {local_configuration_root}
    return translate_paths(filter_paths, original_directory)


class UpdatedPaths(NamedTuple):
    updated_paths: List[str]
    deleted_paths: List[str]

    def is_empty(self) -> bool:
        return not self.updated_paths and not self.deleted_paths


class AnalysisDirectory:
    def __init__(
        self,
        path: str,
        filter_paths: Optional[Set[str]] = None,
        search_path: Optional[List[str]] = None,
    ) -> None:
        self._path = path
        self._filter_paths: Set[str] = filter_paths or set()
        self._search_path_directories: List[str] = self._get_search_path_directories(
            search_path or []
        )

    def get_root(self) -> str:
        return self._path

    def get_filter_roots(self) -> Set[str]:
        current_project_directories = self._filter_paths or {self.get_root()}
        return {
            translate_path(os.getcwd(), filter_root)
            for filter_root in current_project_directories
        }

    def prepare(self) -> None:
        pass

    def compute_symbolic_links(self) -> Dict[str, str]:
        return {}

    def process_updated_files(self, paths: List[str]) -> UpdatedPaths:
        """
            Process a list of paths which were added/removed/updated, making any
            necessary changes to the directory:
                - For an AnalysisDirectory, nothing needs to be changed, since
                  the mapping from source file to analysis file is 1:1.
                - For a SharedAnalysisDirectory, the symbolic links (as well as
                  the reverse-mapping we track) need to be updated to account for
                  new and deleted files.

            Return a list of files (corresponding to the given paths) that Pyre
            should be tracking.
        """
        deleted_paths = [path for path in paths if not os.path.isfile(path)]
        tracked_paths = [
            path
            for path in paths
            if self._is_tracked(path) and path not in deleted_paths
        ]
        tracked_paths.extend(deleted_paths)
        return UpdatedPaths(updated_paths=tracked_paths, deleted_paths=deleted_paths)

    def cleanup(self) -> None:
        pass

    @staticmethod
    def _get_search_path_directories(search_path: List[str]) -> List[str]:
        return [os.path.abspath(os.path.join(*path.split("$"))) for path in search_path]

    @property
    @functools.lru_cache(1)
    def _tracked_directories(self) -> List[str]:
        return [os.path.abspath(self.get_root()), *self._search_path_directories]

    def _is_tracked(self, path: str) -> bool:
        return any(
            is_parent(directory, path) for directory in self._tracked_directories
        )

    def _is_in_search_path(self, path: str) -> bool:
        return any(
            is_parent(directory, path) for directory in self._search_path_directories
        )

    def acquire_shared_reader_lock(self) -> ContextManager[Optional[int]]:
        return do_nothing()

    def acquire_writer_lock(self) -> ContextManager[Optional[int]]:
        return do_nothing()


class SharedAnalysisDirectory(AnalysisDirectory):
    def __init__(
        self,
        source_directories: List[str],
        targets: List[str],
        project_root: str,
        original_directory: Optional[str] = None,
        filter_paths: Optional[Set[str]] = None,
        local_configuration_root: Optional[str] = None,
        extensions: Optional[List[str]] = None,
        search_path: Optional[List[str]] = None,
        isolate: bool = False,
        buck_builder: Optional[BuckBuilder] = None,
        configuration: Optional[Configuration] = None,
    ) -> None:
        self._source_directories: Set[str] = set(source_directories)
        self._targets: Set[str] = set(targets)
        self._original_directory = original_directory
        self._project_root = project_root
        self._filter_paths: Set[str] = filter_paths or set()
        self._local_configuration_root = local_configuration_root
        self._extensions: Set[str] = set(extensions or []) | {"py", "pyi"}
        self._search_path_directories: List[str] = self._get_search_path_directories(
            search_path or []
        )
        self._isolate = isolate
        self._buck_builder: BuckBuilder = buck_builder or buck.SimpleBuckBuilder()

        # Mapping from source files in the project root to symbolic links in the
        # analysis directory.
        self._symbolic_links: Dict[str, str] = {}

        self._configuration = configuration
        self._last_singly_deleted_path_and_link: Optional[Tuple[str, str]] = None

    def get_scratch_directory(self) -> str:
        try:
            return (
                subprocess.check_output(["mkscratch", "path", "--subdir", "pyre"])
                .decode("utf-8")
                .strip()
            )
        except Exception:
            return os.path.join(self._project_root, ".pyre")

    @functools.lru_cache(1)
    def get_root(self) -> str:
        path_to_root = self._local_configuration_root or "shared_analysis_directory"
        suffix = "_{}".format(str(os.getpid())) if self._isolate else ""
        return os.path.join(
            self.get_scratch_directory(), "{}{}".format(path_to_root, suffix)
        )

    def get_filter_roots(self) -> Set[str]:
        current_project_directories = self._filter_paths or {self._project_root}
        if len(self._targets) == 0:
            return {
                translate_path(os.getcwd(), filter_root)
                for filter_root in current_project_directories
            }
        else:
            buck_root = find_buck_root(self._project_root)
            if buck_root is None:
                raise EnvironmentException(
                    "Cannot find buck root when constructing filter directories"
                )
            return {
                translate_path(buck_root, filter_root)
                for filter_root in current_project_directories
            }

    # Exposed for testing.
    def _resolve_source_directories(self) -> None:
        if self._targets:
            new_source_directories = self._buck_builder.build(self._targets)
            original_directory = self._original_directory
            if original_directory is not None:
                new_source_directories = translate_paths(
                    # pyre-fixme[6]: Expected `Set[str]` for 1st positional only parameter
                    # to call `translate_paths` but got `typing.Iterable[str]`.
                    new_source_directories,
                    original_directory,
                )
            self._source_directories.update(new_source_directories)

        if len(self._source_directories) == 0:
            message = """
            No targets or source directories to analyze.
            Did you mean to run the command within a Pyre project?
            (A Pyre project is a directory with a `.pyre_configuration.local` file.
            If it is in directory foo/, you can also run `pyre -l foo`.)
            See `pyre --help` for more details.
            """
            raise NotWithinLocalConfigurationException(textwrap.dedent(message).strip())

    def prepare(self) -> None:
        start = time()
        root = self.get_root()
        LOG.info("Constructing shared directory `%s`", root)

        self._resolve_source_directories()

        try:
            os.makedirs(root)
        except OSError:
            pass  # Swallow.

        lock = os.path.join(root, ".pyre.lock")
        with acquire_lock_if_needed(lock, blocking=True, needed=not self._isolate):
            self._clear()
            self._merge()
            LOG.log(
                log.PERFORMANCE, "Merged analysis directories in %fs", time() - start
            )
        self._symbolic_links.update(self.compute_symbolic_links())

    def rebuild(self) -> None:
        start = time()

        root = self.get_root()
        LOG.info("Updating shared directory `%s`", root)

        self._resolve_source_directories()

        with acquire_lock_if_needed(
            os.path.join(root, ".pyre.lock"), blocking=True, needed=not self._isolate
        ):
            all_paths = {}
            for source_directory in self._source_directories:
                self._merge_into_paths(source_directory, all_paths)
            for relative_path, project_path in all_paths.items():
                scratch_path = os.path.join(root, relative_path)
                if os.path.realpath(scratch_path) != project_path:
                    add_symbolic_link(scratch_path, project_path)
            for scratch_path in self._symbolic_links.values():
                if not os.path.exists(scratch_path):
                    os.remove(scratch_path)
            LOG.log(log.PERFORMANCE, "Updated shared directory in %fs", time() - start)
        self._symbolic_links = self.compute_symbolic_links()

    def compute_symbolic_links(self) -> Dict[str, str]:
        return _compute_symbolic_link_mapping(self.get_root(), self._extensions)

    @staticmethod
    def should_rebuild(
        updated_tracked_paths: List[str], new_paths: List[str], deleted_paths: List[str]
    ) -> bool:
        if any(
            Path(path).suffix == ".thrift" or Path(path).name == "TARGETS"
            for path in chain(updated_tracked_paths, new_paths, deleted_paths)
        ):
            return True
        return (
            len(updated_tracked_paths) >= REBUILD_THRESHOLD_FOR_UPDATED_PATHS
            or len(new_paths) + len(deleted_paths)
            >= REBUILD_THRESHOLD_FOR_NEW_OR_DELETED_PATHS
        )

    def _notify_about_rebuild(self, is_start_message: bool = True) -> None:
        configuration = self._configuration
        if configuration is None:
            return

        LOG.info(
            "Notifying server of %s of rebuild.", "start" if is_start_message else "end"
        )
        if is_start_message:
            message = (
                "Pyre is rebuilding because a significant number of files were "
                "changed. Your results may be outdated until this is finished."
            )
            short_message = "Rebuilding..."
            message_type = json_rpc.LanguageServerMessageType.WARNING.value
        else:
            message = "Done rebuilding."
            short_message = "Done rebuilding."
            message_type = json_rpc.LanguageServerMessageType.INFORMATION.value
        show_status_message = json_rpc.Request(
            method="window/showStatus",
            parameters={
                "message": message,
                "shortMessage": short_message,
                "type": message_type,
                "actions": [],
                "progress": {
                    "numerator": DONT_CARE_PROGRESS_VALUE,
                    "denominator": DONT_CARE_PROGRESS_VALUE,
                },
            },
        )
        try:
            with SocketConnection(configuration.log_directory) as socket_connection:
                socket_connection.perform_handshake(configuration.version_hash)
                socket_connection.send(show_status_message)
        except (
            SocketException,
            ResourceWarning,
            json_rpc.JSONRPCException,
        ) as exception:
            LOG.error("Error while communicating with server: %s", str(exception))

    def _cache_last_deleted_link(
        self,
        deleted_paths: List[str],
        deleted_scratch_paths: List[str],
        tracked_search_paths: List[str],
    ) -> None:
        """Cache the current deleted path and link if there is only one updated
        path and it is deleted. This will help avoid a costly buck query for
        Vim's behavior of deletion followed by addition."""
        if (
            len(deleted_scratch_paths) == 1
            and tracked_search_paths == deleted_scratch_paths
        ):
            self._last_singly_deleted_path_and_link = (
                deleted_paths[0],
                deleted_scratch_paths[0],
            )
            LOG.debug(
                "Caching singly-deleted file and link: %s",
                self._last_singly_deleted_path_and_link,
            )
        else:
            self._last_singly_deleted_path_and_link = None

    def _fetch_cached_absolute_link_map(
        self, new_paths: List[str], deleted_paths: List[str]
    ) -> Optional[Dict[str, str]]:
        """Fetch a cached absolute link map if it exists.

        Editing a file in Vim leads to a "deletion" and "addition" for the
        edited file. If the user calls `pyre` after we process the deletion but
        before we process the addition, then we get spurious "could not find the
        module" errors. So, cache the link map that the seconds-long `buck
        query` would have returned so that there is less chance of such a
        race."""
        last_singly_deleted_path_and_link = self._last_singly_deleted_path_and_link
        if (
            last_singly_deleted_path_and_link is not None
            and new_paths == [last_singly_deleted_path_and_link[0]]
            and deleted_paths == []
        ):
            return {
                last_singly_deleted_path_and_link[0]: last_singly_deleted_path_and_link[
                    1
                ]
            }
        else:
            return None

    def _process_rebuilt_files(
        self, tracked_paths: List[str], deleted_paths: List[str]
    ) -> Tuple[List[str], List[str]]:
        self._notify_about_rebuild(is_start_message=True)

        old_symbolic_links = self._symbolic_links
        old_paths = set(self._symbolic_links.keys())

        # We need to inform the server of any files updated in place by the
        # rebuild, such as .pyi files from thrift.
        rebuild_start_time = time()

        self.rebuild()
        new_paths = set(self._symbolic_links.keys())

        self._notify_about_rebuild(is_start_message=False)

        newly_created_paths = new_paths - old_paths
        tracked_paths.extend(newly_created_paths)

        # Using the modified time instead of a Watchman `since` query because
        # these files will be in the buck builder cache or in /tmp, and Watchman
        # doesn't track those.
        updated_paths = [
            shared_analysis_path
            for shared_analysis_path, original_path in self._symbolic_links.items()
            if os.path.getmtime(original_path) > rebuild_start_time
        ]
        tracked_paths.extend(updated_paths)

        # Translate the paths here because we need the old symbolic links
        # mapping to get their old scratch path.
        deleted_scratch_paths = [
            old_symbolic_links.get(path, path) for path in old_paths - new_paths
        ]
        return tracked_paths, deleted_scratch_paths

    def _process_new_paths(
        self, new_paths: List[str], tracked_paths: List[str], deleted_paths: List[str]
    ) -> List[str]:
        absolute_link_map = self._fetch_cached_absolute_link_map(
            new_paths, deleted_paths
        )
        if absolute_link_map is None:
            relative_link_map = {}
            try:
                relative_link_map = buck.query_buck_relative_paths(
                    new_paths, self._targets
                )
            except buck.BuckException as error:
                LOG.error("Exception occurred when querying buck: %s", error)
                LOG.error("No new paths will be added to the analysis directory.")

            absolute_link_map = {
                path: os.path.join(self.get_root(), relative_link)
                for path, relative_link in relative_link_map.items()
            }
        tracked_paths.extend(absolute_link_map.keys())
        for path, absolute_link in absolute_link_map.items():
            try:
                add_symbolic_link(absolute_link, path)
                self._symbolic_links[path] = absolute_link
            except OSError:
                LOG.warning("Failed to add link at %s.", absolute_link)
        return tracked_paths

    def _process_deleted_paths(
        self, deleted_paths: List[str]
    ) -> Tuple[List[str], List[str]]:
        # Translate the paths here because we need the old symbolic links
        # mapping to get their old scratch path.
        deleted_scratch_paths = [self._symbolic_links[path] for path in deleted_paths]
        for path in deleted_paths:
            link = self._symbolic_links.pop(path, None)
            if link:
                try:
                    _delete_symbolic_link(link)
                except OSError:
                    LOG.warning("Failed to delete link at `%s`.", link)
        return deleted_paths, deleted_scratch_paths

    def _get_new_deleted_and_tracked_paths(
        self, paths: List[str]
    ) -> Tuple[List[str], List[str], List[str]]:
        deleted_paths = [
            path
            for path in paths
            if not os.path.isfile(path)
            and (path in self._symbolic_links or self._is_tracked(path))
        ]
        new_paths = [
            path
            for path in paths
            if path not in self._symbolic_links
            and not self._is_in_search_path(path)
            and os.path.isfile(path)
            and is_parent(self._project_root, path)
        ]
        tracked_paths = [
            path
            for path in paths
            if path not in new_paths
            and path not in deleted_paths
            and (path in self._symbolic_links or self._is_tracked(path))
        ]
        return new_paths, deleted_paths, tracked_paths

    def _process_updated_files(self, paths: List[str]) -> UpdatedPaths:
        """Update the analysis directory for any new or deleted files.
        Rebuild the directory using buck if needed.
        Return the updated and deleted paths."""
        deleted_scratch_paths = []
        (
            new_paths,
            deleted_paths,
            tracked_paths,
        ) = self._get_new_deleted_and_tracked_paths(paths)
        if SharedAnalysisDirectory.should_rebuild(
            tracked_paths, new_paths, deleted_paths
        ):
            tracked_paths, deleted_scratch_paths = self._process_rebuilt_files(
                tracked_paths, deleted_paths
            )
        elif new_paths or deleted_paths:
            if new_paths:
                LOG.info("Detected new paths: %s.", ",".join(new_paths))
                tracked_paths = self._process_new_paths(
                    new_paths, tracked_paths, deleted_paths
                )
            if deleted_paths:
                LOG.info("Detected deleted paths: `%s`.", "`,`".join(deleted_paths))
                deleted_paths, deleted_scratch_paths = self._process_deleted_paths(
                    deleted_paths
                )

        tracked_scratch_paths = [
            self._symbolic_links.get(path, path) for path in tracked_paths
        ] + deleted_scratch_paths

        self._cache_last_deleted_link(
            deleted_paths, deleted_scratch_paths, tracked_scratch_paths
        )

        # The server expects scratch paths.
        return UpdatedPaths(
            updated_paths=tracked_scratch_paths, deleted_paths=deleted_scratch_paths
        )

    def process_updated_files(self, paths: List[str]) -> UpdatedPaths:
        with self.acquire_writer_lock():
            return self._process_updated_files(paths)

    def _directories_to_clean_up(self) -> List[str]:
        if not self._isolate:
            return []

        if not isinstance(self._buck_builder, buck.FastBuckBuilder):
            return [self.get_root()]

        project_name = _get_project_name(self._isolate, self._local_configuration_root)
        buck_builder_cache_path = os.path.join(
            self.get_scratch_directory(), f"{BUCK_BUILDER_CACHE_PREFIX}_{project_name}"
        )
        return [self.get_root(), buck_builder_cache_path]

    def cleanup(self) -> None:
        try:
            for directory in self._directories_to_clean_up():
                shutil.rmtree(directory)
        except Exception as exception:
            LOG.debug(f"Could not clean up analysis directory: {exception}")

    def _clear(self) -> None:
        root = self.get_root()
        for path in os.listdir(root):
            if path.startswith(".pyre"):
                continue

            path = os.path.join(root, path)
            remove_if_exists(path)

    def _merge(self) -> None:
        root = self.get_root()

        all_paths = {}
        for source_directory in self._source_directories:
            self._merge_into_paths(source_directory, all_paths)
        for relative, original in all_paths.items():
            merged = os.path.join(root, relative)
            add_symbolic_link(merged, original)

    # Exposed for testing.
    def _merge_into_paths(
        self, source_directory: str, all_paths: Dict[str, str]
    ) -> None:
        paths = find_python_paths(root=source_directory)
        for path in paths:
            relative = os.path.relpath(path, source_directory)
            if not path:
                continue
            # don't bother stat'ing paths that are already in the analysis directory.
            if relative in all_paths:
                continue
            try:
                absolute = os.path.realpath(path)
                # Don't merge symlinked directories.
                if not os.path.isfile(absolute):
                    continue
                if relative.endswith("__init__.py") and is_empty(absolute):
                    # Don't let empty __init__.py files override legitimate files.
                    continue
                all_paths[relative] = absolute
            except FileNotFoundError:
                continue

    def _reader_writer_lock_path(self) -> str:
        return os.path.join(self.get_root(), READER_WRITER_LOCK)

    def acquire_shared_reader_lock(self) -> ContextManager[Optional[int]]:
        return acquire_lock(
            self._reader_writer_lock_path(), blocking=True, is_shared_reader=True
        )

    def acquire_writer_lock(self) -> ContextManager[Optional[int]]:
        return acquire_lock(
            self._reader_writer_lock_path(), blocking=True, is_shared_reader=False
        )


def _get_project_name(
    isolate_per_process: bool, relative_local_root: Optional[str]
) -> Optional[str]:
    if isolate_per_process:
        return f"isolated_{os.getpid()}"
    if relative_local_root:
        return "_".join(Path(relative_local_root).parts)
    return None


def resolve_analysis_directory(
    source_directories: List[str],
    targets: List[str],
    configuration: Configuration,
    original_directory: str,
    project_root: str,
    filter_directory: Optional[str],
    use_buck_builder: Optional[bool],
    debug: bool,
    buck_mode: Optional[str],
    isolate: bool = False,
    relative_local_root: Optional[str] = None,
) -> AnalysisDirectory:
    # Generate filter directories based on command-line input.
    if filter_directory:
        filter_paths = {filter_directory}
    else:
        filter_paths = _resolve_filter_paths(
            source_directories, targets, configuration, original_directory
        )

    # Only read from the configuration if no explicit targets are passed in.
    if not source_directories and not targets:
        source_directories = configuration.source_directories
        targets = configuration.targets
    else:
        source_directories = source_directories or []
        targets = targets or []
        if targets:
            configuration_name = LOCAL_CONFIGURATION_FILE
            command = "pyre init --local"
        else:
            configuration_name = CONFIGURATION_FILE
            command = "pyre init"
        LOG.warning(
            "Setting up a `%s` with `%s` may reduce overhead.",
            configuration_name,
            command,
        )

    local_configuration_root = configuration.local_root
    if local_configuration_root:
        local_configuration_root = os.path.relpath(
            local_configuration_root, project_root
        )

    use_buck_builder: bool = (
        use_buck_builder
        if use_buck_builder is not None
        else configuration.use_buck_builder
    )

    if len(source_directories) == 1 and len(targets) == 0:
        analysis_directory = AnalysisDirectory(
            source_directories.pop(),
            filter_paths=filter_paths,
            search_path=configuration.search_path,
        )
    else:
        if use_buck_builder:
            buck_root = find_buck_root(project_root)
            if not buck_root:
                raise EnvironmentException(
                    f"No Buck configuration at `{project_root}` or any of its ancestors."
                )
            project_name = _get_project_name(
                isolate_per_process=isolate, relative_local_root=relative_local_root
            )
            buck_builder = buck.FastBuckBuilder(
                buck_root=buck_root,
                buck_builder_binary=configuration.buck_builder_binary,
                debug_mode=debug,
                buck_mode=buck_mode,
                project_name=project_name,
            )
        else:
            buck_builder = buck.SimpleBuckBuilder()

        analysis_directory = SharedAnalysisDirectory(
            source_directories=source_directories,
            targets=targets,
            buck_builder=buck_builder,
            original_directory=original_directory,
            project_root=project_root,
            filter_paths=filter_paths,
            local_configuration_root=local_configuration_root,
            extensions=configuration.extensions,
            search_path=configuration.search_path,
            isolate=isolate,
            configuration=configuration,
        )
    return analysis_directory



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import json
import re
import select
import sys
import time
from typing import List, Optional

from ..analysis_directory import AnalysisDirectory
from ..configuration import Configuration
from .command import Command, CommandArguments, IncrementalStyle
from .start import Start


class Persistent(Command):
    NAME = "persistent"

    def __init__(
        self,
        command_arguments: CommandArguments,
        original_directory: str,
        *,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
        no_watchman: bool,
    ) -> None:
        super(Persistent, self).__init__(
            command_arguments, original_directory, configuration, analysis_directory
        )
        self._no_watchman: bool = no_watchman

    @staticmethod
    def from_arguments(
        arguments: argparse.Namespace,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> "Persistent":
        return Persistent(
            CommandArguments.from_arguments(arguments),
            original_directory,
            configuration=configuration,
            analysis_directory=analysis_directory,
            no_watchman=arguments.no_watchman,
        )

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        persistent = parser.add_parser(
            cls.NAME,
            epilog="""
            Entry point for IDE integration to Pyre. Communicates with a
            Pyre server using the Language Server Protocol, accepts input from stdin and
            writing diagnostics and responses from the Pyre server to stdout.
            """,
        )
        persistent.set_defaults(command=cls.from_arguments, noninteractive=True)
        persistent.add_argument(
            "--no-watchman",
            action="store_true",
            help="Do not spawn a watchman client in the background.",
        )

    def _run(self) -> None:
        Start(
            self._command_arguments,
            self._original_directory,
            terminal=False,
            store_type_check_resolution=False,
            use_watchman=not self._no_watchman,
            incremental_style=IncrementalStyle.FINE_GRAINED,
            configuration=self._configuration,
            analysis_directory=self._analysis_directory,
        ).run()

        self._call_client(command=self.NAME, capture_output=False).check()

    def _flags(self) -> List[str]:
        flags = [
            "-log-identifier",
            '"{}"'.format(self._analysis_directory.get_root()),
            "-expected-binary-version",
            self._configuration.version_hash,
        ]
        if self._configuration.autocomplete:
            flags.append("-autocomplete")

        if self._log_directory:
            flags.extend(["-log-directory", self._log_directory])
        flags += self._feature_flags()
        return flags

    @classmethod
    def _initialize_response(cls, request_id: int) -> str:
        response = json.dumps(
            {"jsonrpc": "2.0", "id": request_id, "result": {"capabilities": {}}},
            sort_keys=True,
        )
        return "Content-Length: {}\r\n\r\n{}\r\n".format(len(response), response)

    @classmethod
    def run_null_server(cls, timeout: Optional[int] = None) -> None:
        to_read, _, _ = select.select([sys.stdin], [], [], 3.0)
        request_id = 0
        if to_read:
            standard_input = to_read[0]
            # Read content of the form Content-Length:n\r\n\r\n{jsonmessage}
            line = standard_input.readline()
            match = re.match(r"Content-Length: (?P<bytes>[0-9]+)", line)
            if match:
                length = int(match.group("bytes"))
                serialized_json = standard_input.read(length)
                try:
                    parsed = json.loads(serialized_json)
                    request_id = parsed["id"]
                # This is a catch-all to ensure that the null server always
                # gets spawned.
                except Exception:
                    pass

        sys.stdout.write(cls._initialize_response(request_id))
        sys.stdout.flush()
        start_time = int(time.time())
        while True:
            if timeout is not None and timeout <= (int(time.time()) - start_time):
                break
            time.sleep(10)



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Type

from .analyze import Analyze as Analyze
from .check import Check as Check
from .deobfuscate import Deobfuscate as Deobfuscate
from .incremental import Incremental as Incremental
from .infer import Infer as Infer
from .initialize import Initialize as Initialize
from .kill import Kill as Kill
from .persistent import Persistent as Persistent
from .profile import Profile as Profile
from .query import Query as Query
from .rage import Rage as Rage
from .reporting import Reporting as Reporting
from .restart import Restart as Restart
from .servers import Servers as Servers
from .start import Start
from .statistics import Statistics as Statistics
from .stop import Stop as Stop


from .command import (  # noqa; noqa; noqa
    ClientException as ClientException,
    Command as Command,
    CommandParser as CommandParser,
    ExitCode as ExitCode,
    ProfileOutput as ProfileOutput,
    typeshed_search_path as typeshed_search_path,
    IncrementalStyle as IncrementalStyle,
)

COMMANDS: List[Type[CommandParser]] = [
    Analyze,
    Check,
    Deobfuscate,
    Incremental,
    Infer,
    Initialize,
    Kill,
    Persistent,
    Profile,
    Query,
    Rage,
    Reporting,
    Restart,
    Servers,
    Start,
    Statistics,
    Stop,
]



# pyre-unsafe

from typing import Any

def date(value, arg: Any = ...) -> Any: ...
def escapejs(value: str) -> str: ...
def linebreaks(value: str, autoescape: bool = ...) -> str: ...
def slugify(value: str) -> str: ...



# pyre-unsafe

from typing import Any, Optional

from django.core.exceptions import ValidationError as ValidationError

class EmailValidator:
    def __call__(self, value: Optional[str]) -> None: ...

class URLValidator:
    def __call__(self, value: Optional[str]) -> None: ...

class RegexValidator(object):
    def __init__(
        self, regex=..., message=..., code=..., inverse_match=..., flags=...
    ) -> None: ...
    def __call__(self, value: Optional[str]) -> None: ...

validate_email: EmailValidator = ...



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

TYPECHECK_ERROR = 102


class AIException(Exception):
    pass


class AIRecoverableException(AIException):
    pass


class AIProcessException(AIRecoverableException):
    def __init__(self, message, error_code):
        super().__init__(message)
        self.error_code = error_code


class ParseTypeException(Exception):
    pass



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from collections import defaultdict
from re import compile
from typing import Any, Dict, List, Pattern, Sequence

import libcst as cst
from libcst.metadata import CodeRange, PositionProvider


class StatisticsCollector(cst.CSTVisitor):
    def build_json(self) -> Dict[str, int]:
        return {}


class AnnotationCountCollector(StatisticsCollector):
    def __init__(
        self,
        return_count: int = 0,
        annotated_return_count: int = 0,
        globals_count: int = 0,
        annotated_globals_count: int = 0,
        parameter_count: int = 0,
        annotated_parameter_count: int = 0,
        attribute_count: int = 0,
        annotated_attribute_count: int = 0,
        partially_annotated_function_count: int = 0,
        fully_annotated_function_count: int = 0,
    ) -> None:
        self.return_count = return_count
        self.annotated_return_count = annotated_return_count
        self.globals_count = globals_count
        self.annotated_globals_count = annotated_globals_count
        self.parameter_count = parameter_count
        self.annotated_parameter_count = annotated_parameter_count
        self.attribute_count = attribute_count
        self.annotated_attribute_count = annotated_attribute_count
        self.partially_annotated_function_count = partially_annotated_function_count
        self.fully_annotated_function_count = fully_annotated_function_count
        self.in_class_definition = False
        self.in_function_definition = False
        self.is_static_function = False

    def build_json(self) -> Dict[str, int]:
        return {
            "return_count": self.return_count,
            "annotated_return_count": self.annotated_return_count,
            "globals_count": self.globals_count,
            "annotated_globals_count": self.annotated_globals_count,
            "parameter_count": self.parameter_count,
            "annotated_parameter_count": self.annotated_parameter_count,
            "attribute_count": self.attribute_count,
            "annotated_attribute_count": self.annotated_attribute_count,
            "partially_annotated_function_count": (
                self.partially_annotated_function_count
            ),
            "fully_annotated_function_count": self.fully_annotated_function_count,
        }

    def _is_self_or_cls(self, index: int) -> bool:
        return index == 0 and self.in_class_definition and not self.is_static_function

    def _check_parameter_annotations(self, parameters: Sequence[cst.Param]) -> int:
        annotated_parameter_count = 0
        for index, parameter in enumerate(parameters):
            self.parameter_count += 1
            annotation = parameter.annotation
            if annotation is not None or self._is_self_or_cls(index):
                annotated_parameter_count += 1
        self.annotated_parameter_count += annotated_parameter_count
        return annotated_parameter_count

    def visit_FunctionDef(self, node: cst.FunctionDef) -> None:
        for decorator in node.decorators:
            decorator_node = decorator.decorator
            if isinstance(decorator_node, cst.Name):
                if decorator_node.value == "staticmethod":
                    self.is_static_function = True
        self.in_function_definition = True

        self.return_count += 1
        return_is_annotated = node.returns is not None
        if return_is_annotated:
            self.annotated_return_count += 1

        annotated_parameters = self._check_parameter_annotations(node.params.params)

        if return_is_annotated and (annotated_parameters == len(node.params.params)):
            self.fully_annotated_function_count += 1
        elif return_is_annotated or annotated_parameters > 0:
            self.partially_annotated_function_count += 1

    def leave_FunctionDef(self, original_node: cst.FunctionDef) -> None:
        self.in_function_definition = False
        self.is_static_function = False

    def visit_Assign(self, node: cst.Assign) -> None:
        if self.in_function_definition:
            return
        if self.in_class_definition:
            self.attribute_count += 1
        else:
            self.globals_count += 1

    def visit_AnnAssign(self, node: cst.AnnAssign) -> None:
        if self.in_function_definition:
            return
        if self.in_class_definition:
            self.attribute_count += 1
            self.annotated_attribute_count += 1
        else:
            self.globals_count += 1
            self.annotated_globals_count += 1

    def visit_ClassDef(self, node: cst.ClassDef) -> None:
        self.in_class_definition = True

    def leave_ClassDef(self, original_node: cst.ClassDef) -> None:
        self.in_class_definition = False


class CountCollector(StatisticsCollector):
    def __init__(self, regex: str) -> None:
        self.counts: Dict[str, int] = defaultdict(int)
        self.regex: Pattern[str] = compile(regex)

    def visit_Comment(self, node: cst.Comment) -> None:
        match = self.regex.match(node.value)
        if match:
            self.counts[match.group(1)] += 1

    def build_json(self) -> Dict[str, int]:
        return dict(self.counts)


class FixmeCountCollector(CountCollector):
    def __init__(self) -> None:
        super().__init__(r"# pyre-fixme\[(\d*)\]:")


class IgnoreCountCollector(CountCollector):
    def __init__(self) -> None:
        super().__init__(r"# pyre-ignore\[(\d*)\]:")


class StrictCountCollector(StatisticsCollector):
    def __init__(self, strict_by_default: bool) -> None:
        self.is_strict: bool = False
        self.is_unsafe: bool = False
        self.strict_count: int = 0
        self.unsafe_count: int = 0
        self.strict_by_default: bool = strict_by_default
        self.unsafe_regex: Pattern[str] = compile(r"# pyre-unsafe")
        self.strict_regex: Pattern[str] = compile(r"# pyre-strict")

    def is_unsafe_module(self) -> bool:
        if self.is_unsafe:
            return True
        elif self.is_strict or self.strict_by_default:
            return False
        return True

    def visit_Module(self, node: cst.Module) -> None:
        self.is_strict = False
        self.is_unsafe = False

    def visit_Comment(self, node: cst.Comment) -> None:
        strict_match = self.strict_regex.match(node.value)
        if strict_match:
            self.is_strict = True
        unsafe_match = self.unsafe_regex.match(node.value)
        if unsafe_match:
            self.is_unsafe = True

    def leave_Module(self, original_node: cst.Module) -> None:
        if self.is_unsafe_module():
            self.unsafe_count += 1
        else:
            self.strict_count += 1

    def build_json(self) -> Dict[str, int]:
        return {"unsafe_count": self.unsafe_count, "strict_count": self.strict_count}


class CodeQualityIssue:
    def __init__(
        self, code_range: CodeRange, path: str, category: str, message: str
    ) -> None:
        self.category: str = category
        self.detail_message: str = message
        self.line: int = code_range.start.line
        self.line_end: int = code_range.end.line
        self.column: int = code_range.start.column
        self.column_end: int = code_range.end.column
        self.path: str = path

    def build_json(self) -> Dict[str, Any]:
        return {
            "category": self.category,
            "detail_message": self.detail_message,
            "line": self.line,
            "line_end": self.line_end,
            "column": self.column,
            "column_end": self.column_end,
            "path": self.path,
        }


class FunctionsCollector(cst.CSTVisitor):
    METADATA_DEPENDENCIES = (PositionProvider,)
    path: str = ""

    def __init__(self) -> None:
        self.issues: List[CodeQualityIssue] = []

    def visit_FunctionDef(self, node: cst.FunctionDef) -> None:
        return_is_annotated = node.returns is not None
        if not return_is_annotated:
            code_range = self.get_metadata(PositionProvider, node)
            issue = CodeQualityIssue(
                code_range,
                self.path,
                "PYRE_MISSING_ANNOTATIONS",
                "This function is missing a return annotation. \
                Bodies of unannotated functions are not typechecked by Pyre.",
            )
            self.issues.append(issue)


class StrictIssueCollector(StrictCountCollector):
    METADATA_DEPENDENCIES = (PositionProvider,)
    path: str = ""
    issues: List[CodeQualityIssue] = []

    def _create_issue(self, node: cst.Module) -> None:
        file_range = self.get_metadata(PositionProvider, node)
        code_range = CodeRange(start=file_range.start, end=file_range.start)
        issue = CodeQualityIssue(
            code_range, self.path, "PYRE_STRICT", "Unsafe Pyre file."
        )
        self.issues.append(issue)

    def leave_Module(self, original_node: cst.Module) -> None:
        if self.is_unsafe_module():
            self._create_issue(original_node)



# pyre-unsafe

from django.utils.safestring import SafeText

def escape(text: str) -> str: ...
def conditional_escape(text: str) -> str: ...
def format_html(format_string, *args, **kwargs) -> SafeText: ...
def format_html_join(sep, format_string, args_generator) -> SafeText: ...



# Copyright (c) 2018-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


class EnvironmentException(Exception):
    pass



import unittest.mock
from typing import Any, Callable, Optional, TypeVar, Union, overload

from ._version import version as version

_T = TypeVar("_T")
_MockModule = Any  # usually unittest.mock
_Config = Any  # pytest.Config

__version__ = version

def _get_mock_module(config: _Config) -> _MockModule: ...

class MockFixture:
    mock_module: _MockModule
    patch: MockFixture._Patcher  # google/pytype#611
    # The following aliases don't work due to google/pytype#612
    Mock: Any  # actually unittest.mock.Mock
    MagicMock: Any  # actually unittest.mock.MagicMock
    NonCallableMock: Any  # actually unittest.mock.NonCallableMock
    PropertyMock: Any  # actually unittest.mock.PropertyMock
    call: Any  # actually unittest.mock.call
    ANY: Any  # actually unittest.mock.ANY
    DEFAULT: Any  # actually unittest.mock.DEFAULT
    create_autospec: Any  # actually unittest.mock.create_autospec
    sentinel: Any  # actually unittest.mock.sentinel
    mock_open: Any  # actually unittest.mock.mock_open
    def __init__(self, config: _Config) -> None: ...
    def resetall(self) -> None: ...
    def stopall(self) -> None: ...
    def spy(self, obj: object, name: str) -> unittest.mock.MagicMock: ...
    def stub(self, name: Optional[str] = ...) -> unittest.mock.MagicMock: ...
    class _Patcher:
        mock_module: _MockModule
        def object(
            self,
            target: Any,
            attribute: str,
            new: Optional[Any] = ...,
            spec: Optional[Any] = ...,
            create: bool = ...,
            spec_set: Optional[Any] = ...,
            autospec: Optional[Any] = ...,
            new_callable: Optional[Any] = ...,
            **kwargs: Any,
        ) -> Any: ...
        def multiple(
            self,
            target: Any,
            spec: Optional[Any] = ...,
            create: bool = ...,
            spec_set: Optional[Any] = ...,
            autospec: Optional[Any] = ...,
            new_callable: Optional[Any] = ...,
            **kwargs: Any,
        ) -> Any: ...
        def dict(self, in_dict: Any, values: Any = ..., clear: bool = ..., **kwargs: Any) -> Any: ...
        @overload
        def __call__(
            self,
            target: Any,
            new: None = ...,
            spec: Optional[Any] = ...,
            create: bool = ...,
            spec_set: Optional[Any] = ...,
            autospec: Optional[Any] = ...,
            new_callable: Optional[Any] = ...,
            **kwargs: Any,
        ) -> unittest.mock.MagicMock: ...
        @overload
        def __call__(
            self,
            target: Any,
            new: _T,
            spec: Optional[Any] = ...,
            create: bool = ...,
            spec_set: Optional[Any] = ...,
            autospec: Optional[Any] = ...,
            new_callable: Optional[Any] = ...,
            **kwargs: Any,
        ) -> _T: ...

mocker: Any
class_mocker: Any
module_mocker: Any
package_mocker: Any
session_mocker: Any

def assert_wrapper(__wrapped_mock_method__: Callable[..., Any], *args: Any, **kwargs: Any) -> None: ...
def wrap_assert_not_called(*args: Any, **kwargs: Any) -> None: ...
def wrap_assert_called_with(*args: Any, **kwargs: Any) -> None: ...
def wrap_assert_called_once(*args: Any, **kwargs: Any) -> None: ...
def wrap_assert_called_once_with(*args: Any, **kwargs: Any) -> None: ...
def wrap_assert_has_calls(*args: Any, **kwargs: Any) -> None: ...
def wrap_assert_any_call(*args: Any, **kwargs: Any) -> None: ...
def wrap_assert_called(*args: Any, **kwargs: Any) -> None: ...
def wrap_assert_methods(config: _Config) -> None: ...
def unwrap_assert_methods() -> None: ...
def pytest_addoption(parser) -> None: ...
def parse_ini_boolean(value: Union[bool, str]) -> bool: ...
def pytest_configure(config: _Config) -> None: ...



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from pathlib import Path
from typing import Any, Dict, List, Optional

from .analysis_directory import AnalysisDirectory
from .commands import stop
from .commands.command import CommandArguments
from .configuration import Configuration
from .find_directories import CONFIGURATION_FILE, LOCAL_CONFIGURATION_FILE

# We use the `LOG` from watchman due to its better formatting in log files
from .watchman import LOG, Subscriber, Subscription


class ConfigurationMonitor(Subscriber):
    """
        The ConfigurationMonitor watches only for .pyre_configuration(.local)
        files, and will kill the corresponding server when a configuration changes.
        Logs are found in
        .pyre/<local-root>/configuration_monitor/configuration_monitor.log
        To kill a monitor, get pid from
        .pyre/<local-root>/configuration_monitor/configuration_monitor.pid
        and kill <pid>.
    """

    def __init__(
        self,
        command_arguments: CommandArguments,
        configuration: Configuration,
        analysis_directory: AnalysisDirectory,
        project_root: str,
        original_directory: str,
        local_configuration_root: Optional[str],
        other_critical_files: List[str],
    ) -> None:
        super(ConfigurationMonitor, self).__init__(self.base_path(configuration))
        self._command_arguments = command_arguments
        self._configuration = configuration
        self._analysis_directory = analysis_directory
        self._project_root_path: Path = Path(project_root).resolve()
        self._original_directory = original_directory
        self._local_configuration_root = local_configuration_root
        self._other_critical_files = other_critical_files

    NAME = "configuration_monitor"

    @property
    def _name(self) -> str:
        return ConfigurationMonitor.NAME

    @staticmethod
    def base_path(configuration: Configuration) -> str:
        return os.path.join(configuration.log_directory, ConfigurationMonitor.NAME)

    @property
    def _subscriptions(self) -> List[Subscription]:
        roots = self._watchman_client.query("watch-list")["roots"]
        LOG.debug(f"Watchman roots: {roots}")
        for root in roots:
            root_path = Path(root).resolve()
            if root_path in (self._project_root_path, *self._project_root_path.parents):
                name = "pyre_monitor_{}".format(os.path.basename(root))
                subscription = {
                    "empty_on_fresh_instance": True,
                    "expression": [
                        "allof",
                        ["type", "f"],
                        [
                            "anyof",
                            ["suffix", "pyre_configuration.local"],
                            ["suffix", "pyre_configuration"],
                            *[
                                ["match", critical]
                                for critical in self._other_critical_files
                            ],
                        ],
                    ],
                    "fields": ["name"],
                }
                subscription = Subscription(root, name, subscription)
                LOG.debug(f"Configuration monitor subscription: {subscription}")
                return [subscription]
        LOG.debug("Configuration monitor is not subscribed to any paths.")
        return []

    def _stop(self) -> None:
        stop.Stop(
            self._command_arguments,
            self._original_directory,
            configuration=self._configuration,
            analysis_directory=self._analysis_directory,
        ).run()

    def _handle_response(self, response: Dict[str, Any]) -> None:
        watchman_root = response.get("root", None)
        paths = response.get("files", None)
        if watchman_root is None or paths is None:
            # Skip the Watchman response because it does not have both `root`
            # and `files`.
            return

        absolute_paths = [Path(watchman_root, path).resolve() for path in paths]
        LOG.info(f"Update to configuration or other critical files at {absolute_paths}")

        root_configuration_path = self._project_root_path / CONFIGURATION_FILE
        local_configuration_root = self._local_configuration_root
        if root_configuration_path in absolute_paths:
            LOG.info("Pyre configuration changed. Stopping pyre server.")
            # TODO(T54088045): Find all local pyre servers running underneath
            # and stop them.
            self._stop()
        elif (
            local_configuration_root is not None
            and Path(local_configuration_root, LOCAL_CONFIGURATION_FILE).resolve()
            in absolute_paths
        ):
            LOG.info("Local configuration changed. Stopping pyre server.")
            self._stop()
        elif any(
            self._project_root_path / critical in absolute_paths
            for critical in self._other_critical_files
        ):
            LOG.info("Critical file changed. Stopping pyre server.")
            self._stop()
        else:
            LOG.info(
                "None of the changed paths correspond to the root "
                "configuration, a local pyre configuration, or any other critical file."
            )



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import ast
import logging
from logging import Logger
from typing import Callable

# pyre-fixme[21]: Could not find name `ListVariadic` in `pyre_extensions`.
from pyre_extensions import ListVariadic
from pyre_extensions.type_variable_operators import Concatenate


Ts = ListVariadic("Ts")


LOG: Logger = logging.getLogger(__name__)


class UnstableAST(Exception):
    pass


def check_stable(input: str, transformed: str) -> None:
    parsed_original = ast.parse(input)
    try:
        parsed_transformed = ast.parse(transformed)
        if ast.dump(parsed_original) != ast.dump(parsed_transformed):
            raise UnstableAST("ASTs differ")
    except SyntaxError:
        raise UnstableAST("Could not parse transformed AST")


def check_stable_transformation(
    transform: "Callable[Concatenate[str, Ts], str]",
) -> "Callable[Concatenate[str, Ts], str]":
    def wrapper(input: str, *args: Ts) -> str:
        transformed = transform(input, *args)
        check_stable(input, transformed)
        return transformed

    return wrapper



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import datetime
import gc
import logging
import sys
import time
from contextlib import contextmanager
from functools import wraps
from typing import Any, Callable, List, Optional


log = logging.getLogger("sapp")


class retryable(object):
    def __init__(
        self, num_tries: int = 1, retryable_exs: Optional[List[Any]] = None
    ) -> None:
        self.num_tries = num_tries
        self.retryable_exs = retryable_exs

    def __call__(self, func):
        @wraps(func)
        def new_func(*args, **kwargs):
            try_num = 1
            while True:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    if self.retryable_exs and type(e) not in self.retryable_exs:
                        raise
                    try_num += 1
                    if try_num > self.num_tries:
                        raise

        new_func.__wrapped__ = func
        return new_func


def log_time(func: Callable[..., Any]) -> Callable[..., Any]:
    """Log the time it takes to run a function. It's sort of like timeit, but
    prettier.
    """

    def wrapper(*args, **kwargs):
        start_time = time.time()
        log.info("%s starting...", func.__name__.title())
        ret = func(*args, **kwargs)
        log.info(
            "%s finished (%s)",
            func.__name__.title(),
            datetime.timedelta(seconds=int(time.time() - start_time)),
        )
        return ret

    return wrapper


class UserError(Exception):
    pass


@contextmanager
def catch_user_error():
    try:
        yield
    except UserError as error:
        print(str(error), file=sys.stderr)


@contextmanager
def catch_keyboard_interrupt():
    try:
        yield
    except KeyboardInterrupt:
        print("\nOperation aborted.", file=sys.stderr)


# For use on enums to alias upper case value.
#
# FLAKE8 does not understand that this is a static property
# flake8: noqa B902
class classproperty(property):
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()



from datetime import timedelta
from typing import Optional, Union

class BadSignature(Exception):
    pass

class SignatureExpired(Exception):
    pass

class Signer:
    def signature(self, value: str) -> str: ...
    def sign(self, value: str) -> str: ...
    def unsign(self, signed_value: str) -> str: ...

class TimestampSigner(Signer):
    # pyre-ignore [14]: original django source violates LSP here
    def unsign(
        self, value: str, max_age: Optional[Union[str, timedelta]] = None
    ) -> str: ...

def get_cookie_signer(salt: str) -> TimestampSigner: ...



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import ast as builtin_ast
import logging
import re
import subprocess
from enum import Enum
from pathlib import Path
from typing import Dict, List, NamedTuple

from ...client.filesystem import get_filesystem
from . import ast


LOG: logging.Logger = logging.getLogger(__name__)


class LocalMode(Enum):
    IGNORE = "pyre-ignore-all-errors"
    UNSAFE = "pyre-unsafe"
    STRICT = "pyre-strict"

    def get_regex(self) -> str:
        return "^[ \t]*# *" + self.value + " *$"

    def get_comment(self) -> str:
        return "# " + self.value


class Target(NamedTuple):
    name: str
    strict: bool
    pyre: bool


class TargetCollector(builtin_ast.NodeVisitor):
    def __init__(self, pyre_only: bool) -> None:
        self._pyre_only: bool = pyre_only
        self._targets: List[Target] = []
        self._contains_strict: bool = False

    def visit_Call(self, node: builtin_ast.Call) -> None:
        target_fields = node.keywords
        check_types = False
        uses_pyre = True
        is_strict = False
        name = None
        for field in target_fields:
            value = field.value
            if field.arg == "name":
                if isinstance(value, builtin_ast.Str):
                    name = value.s
            elif field.arg == "check_types":
                if isinstance(value, builtin_ast.NameConstant):
                    check_types = check_types or value.value
            elif field.arg == "check_types_options":
                if isinstance(value, builtin_ast.Str):
                    uses_pyre = uses_pyre and "mypy" not in value.s.lower()
                    is_strict = is_strict or (uses_pyre and "strict" in value.s.lower())
            elif field.arg == "typing_options":
                if isinstance(value, builtin_ast.Str):
                    is_strict = is_strict or "strict" in value.s.lower()
        if name and check_types and (not self._pyre_only or uses_pyre):
            self._targets.append(Target(name, is_strict, uses_pyre))
        self._contains_strict = self._contains_strict or is_strict

    def result(self) -> List[Target]:
        return self._targets

    def contains_strict(self) -> bool:
        return self._contains_strict


def path_exists(filename: str) -> Path:
    path = Path(filename)
    if not path.exists():
        raise ValueError("No file at {}".format(filename))
    return path


def find_targets(search_root: Path, pyre_only: bool = False) -> Dict[str, List[Target]]:
    LOG.info("Finding typecheck targets in %s", search_root)
    target_files = find_files(search_root, "TARGETS")
    target_names = {}
    total_targets = 0
    for target_file in target_files:
        target_finder = TargetCollector(pyre_only)
        with open(target_file, "r") as source:
            tree = builtin_ast.parse(source.read())
            target_finder.visit(tree)
            targets = target_finder.result()
            if len(targets) > 0:
                target_names[target_file] = targets
                total_targets += len(targets)

    LOG.info(
        "Found {} typecheck targets in {} TARGETS files to analyze".format(
            total_targets, len(target_names)
        )
    )
    return target_names


def add_local_mode(filename: str, mode: LocalMode) -> None:
    LOG.info("Processing `%s`", filename)
    path = Path(filename)
    text = path.read_text()
    if "@" "generated" in text:
        LOG.warning("Attempting to edit generated file %s, skipping.", filename)
        return

    lines = text.split("\n")  # type: List[str]

    # Check if a local mode is already set.
    for line in lines:
        for local_mode in LocalMode:
            if re.match(local_mode.get_regex(), line):
                return

    def is_header(line: str) -> bool:
        is_comment = line.lstrip().startswith("#")
        is_pyre_ignore = (
            re.match("^[ \t]*# *pyre-ignore *$", line)
            or re.match("^[ \t]*# *pyre-fixme *$", line)
            or re.match("^[ \t]*# *type: ignore *$", line)
        )
        return is_comment and not is_pyre_ignore

    # Add local mode.
    new_lines = []
    past_header = False
    for line in lines:
        if not past_header and not is_header(line):
            past_header = True
            if len(new_lines) != 0:
                new_lines.append("")
            new_lines.append(mode.get_comment())
        new_lines.append(line)
    new_text = "\n".join(new_lines)
    ast.check_stable(text, new_text)
    path.write_text(new_text)


def remove_non_pyre_ignores(subdirectory: Path) -> None:
    python_files = [
        str(subdirectory / path)
        for path in get_filesystem().list(str(subdirectory), patterns=[r"**/*.py"])
    ]
    if python_files:
        LOG.info("...cleaning %s python files", len(python_files))
        remove_type_ignore_command = [
            "sed",
            "-i",
            r"s/\s*# \?type: \?ignore$//g",
        ] + python_files
        subprocess.check_output(remove_type_ignore_command)


def find_files(directory: Path, name: str) -> List[str]:
    output = (
        subprocess.check_output(
            ["find", str(directory), "-name", name], stderr=subprocess.DEVNULL
        )
        .decode("utf-8")
        .strip()
    )
    if output == "":
        return []
    files = output.split("\n")
    return [file.strip() for file in files]


def find_directories(directory: Path) -> List[Path]:
    return [path for path in directory.iterdir() if path.is_dir()]



# Copyright (c) 2019-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
import shutil
import subprocess
from logging import Logger
from pathlib import Path
from typing import Optional

from .find_directories import (
    CONFIGURATION_FILE,
    LOCAL_CONFIGURATION_FILE,
    find_project_root,
)


TEST_DIRECTORY = "__TEST_DIRECTORY"
LOG_DIRECTORY: str = ".pyre"

LOG: Logger = logging.getLogger(__name__)


def find_log_directory(
    project_root: str, local_configuration: Optional[str], dot_pyre_directory: str
) -> str:
    """Pyre outputs all logs to a .pyre directory that lives in the project root."""
    log_directory = dot_pyre_directory
    if local_configuration:
        # `log_directory` will never escape `.pyre/` because in `switch_root` we have
        # guaranteed that configurations are never deeper than local configurations
        relative = os.path.relpath(local_configuration, project_root)
        log_directory = os.path.join(log_directory, relative)
    return log_directory


def get_dot_pyre_directory(root_directory: str) -> str:
    dot_pyre_directory = os.path.join(root_directory, LOG_DIRECTORY)
    try:
        scratch_path = (
            subprocess.check_output(["mkscratch", "path", root_directory])
            .decode()
            .strip()
        )
    except Exception as exception:
        LOG.debug("Could not find scratch path because of exception: %s", exception)
    else:
        # Test if we have permission to create a nested directory.
        test_outer_directory = os.path.join(scratch_path, TEST_DIRECTORY)
        test_scratch_directory = os.path.join(test_outer_directory, TEST_DIRECTORY)
        try:
            os.makedirs(test_scratch_directory, exist_ok=True)
        except PermissionError:
            LOG.debug(
                "Did not have permission to write to the dot pyre directory in scratch."
                " Falling back to a dot pyre directory in the project root."
            )
        except Exception as exception:
            LOG.debug(
                f"Could not test if the dot pyre directory is writable:\n{exception}\n"
                "Falling back to a dot pyre directory in the project root."
            )
        else:
            dot_pyre_directory = os.path.join(scratch_path, LOG_DIRECTORY)
        finally:
            try:
                shutil.rmtree(test_outer_directory)
            except Exception:
                pass
    return dot_pyre_directory


def log_directory(
    project_root: str,
    local_root: Optional[str] = None,
    subdirectory: Optional[str] = None,
) -> Path:
    dot_pyre = get_dot_pyre_directory(root_directory=project_root)
    if local_root is None:
        return Path(dot_pyre)
    project_root = find_project_root(original_directory=project_root)
    log_directory = Path(
        find_log_directory(
            dot_pyre_directory=dot_pyre,
            project_root=project_root,
            local_configuration=local_root,
        )
    )
    if subdirectory is None:
        return log_directory
    return log_directory / subdirectory


def get_configuration_value(directory: str, key: str) -> str:
    local_configuration = Path(directory) / LOCAL_CONFIGURATION_FILE
    project_root = Path(find_project_root(original_directory=directory))
    project_configuration = project_root / CONFIGURATION_FILE

    with open(local_configuration) as file:
        local_configuration = json.load(file)

    if local_configuration.get(key):
        return local_configuration.get(key)
    else:
        with open(project_configuration) as file:
            project_configuration = json.load(file)
        return project_configuration.get(key)



from typing import Collection

from django.db.models import Model
from django.db.models.fields import Field

class Collector: ...

def CASCADE(
    collector: Collector, field: Field, sub_objs: Collection[Model], using: str
) -> None: ...
def PROTECT(
    collector: Collector, field: Field, sub_objs: Collection[Model], using: str
) -> None: ...
def RESTRICT(
    collector: Collector, field: Field, sub_objs: Collection[Model], using: str
) -> None: ...
def SET_NULL(
    collector: Collector, field: Field, sub_objs: Collection[Model], using: str
) -> None: ...
def SET_DEFAULT(
    collector: Collector, field: Field, sub_objs: Collection[Model], using: str
) -> None: ...
def DO_NOTHING(
    collector: Collector, field: Field, sub_objs: Collection[Model], using: str
) -> None: ...



# Source: https://github.com/python/cpython/blob/master/Lib/sre_parse.py

import sys
from sre_constants import _NamedIntConstant as _NIC, error as _Error
from typing import Any, Dict, FrozenSet, Iterable, List, Match, Optional, Pattern as _Pattern, Tuple, Union

SPECIAL_CHARS: str
REPEAT_CHARS: str
DIGITS: FrozenSet[str]
OCTDIGITS: FrozenSet[str]
HEXDIGITS: FrozenSet[str]
ASCIILETTERS: FrozenSet[str]
WHITESPACE: FrozenSet[str]
ESCAPES: Dict[str, Tuple[_NIC, int]]
CATEGORIES: Dict[str, Union[Tuple[_NIC, _NIC], Tuple[_NIC, List[Tuple[_NIC, _NIC]]]]]
FLAGS: Dict[str, int]
if sys.version_info >= (3, 6):
    GLOBAL_FLAGS: int
    class Verbose(Exception): ...

class _State:
    flags: int
    groupdict: Dict[str, int]
    groupwidths: List[Optional[int]]
    lookbehindgroups: Optional[int]
    def __init__(self) -> None: ...
    @property
    def groups(self) -> int: ...
    def opengroup(self, name: str = ...) -> int: ...
    def closegroup(self, gid: int, p: SubPattern) -> None: ...
    def checkgroup(self, gid: int) -> bool: ...
    def checklookbehindgroup(self, gid: int, source: Tokenizer) -> None: ...

if sys.version_info >= (3, 8):
    State = _State
else:
    Pattern = _State

_OpSubpatternType = Tuple[Optional[int], int, int, SubPattern]
_OpGroupRefExistsType = Tuple[int, SubPattern, SubPattern]
_OpInType = List[Tuple[_NIC, int]]
_OpBranchType = Tuple[None, List[SubPattern]]
_AvType = Union[_OpInType, _OpBranchType, Iterable[SubPattern], _OpGroupRefExistsType, _OpSubpatternType]
_CodeType = Tuple[_NIC, _AvType]

class SubPattern:
    data: List[_CodeType]
    width: Optional[int]

    if sys.version_info >= (3, 8):
        state: State
        def __init__(self, state: State, data: Optional[List[_CodeType]] = ...) -> None: ...
    else:
        pattern: Pattern
        def __init__(self, pattern: Pattern, data: Optional[List[_CodeType]] = ...) -> None: ...
    def dump(self, level: int = ...) -> None: ...
    def __len__(self) -> int: ...
    def __delitem__(self, index: Union[int, slice]) -> None: ...
    def __getitem__(self, index: Union[int, slice]) -> Union[SubPattern, _CodeType]: ...
    def __setitem__(self, index: Union[int, slice], code: _CodeType) -> None: ...
    def insert(self, index: int, code: _CodeType) -> None: ...
    def append(self, code: _CodeType) -> None: ...
    def getwidth(self) -> int: ...

class Tokenizer:
    istext: bool
    string: Any
    decoded_string: str
    index: int
    next: Optional[str]
    def __init__(self, string: Any) -> None: ...
    def match(self, char: str) -> bool: ...
    def get(self) -> Optional[str]: ...
    def getwhile(self, n: int, charset: Iterable[str]) -> str: ...
    if sys.version_info >= (3, 8):
        def getuntil(self, terminator: str, name: str) -> str: ...
    else:
        def getuntil(self, terminator: str) -> str: ...
    if sys.version_info >= (3, 6):
        @property
        def pos(self) -> int: ...
    def tell(self) -> int: ...
    def seek(self, index: int) -> None: ...
    def error(self, msg: str, offset: int = ...) -> _Error: ...

def fix_flags(src: Union[str, bytes], flags: int) -> int: ...

_TemplateType = Tuple[List[Tuple[int, int]], List[str]]
if sys.version_info >= (3, 8):
    def parse(str: str, flags: int = ..., state: Optional[State] = ...) -> SubPattern: ...
    def parse_template(source: str, state: _Pattern[Any]) -> _TemplateType: ...

else:
    def parse(str: str, flags: int = ..., pattern: Optional[Pattern] = ...) -> SubPattern: ...
    def parse_template(source: str, pattern: _Pattern[Any]) -> _TemplateType: ...

def expand_template(template: _TemplateType, match: Match[Any]) -> str: ...



from typing import Any

from jwt.algorithms import Algorithm

from . import _HashAlg

class ECAlgorithm(Algorithm[Any]):
    SHA256: _HashAlg
    SHA384: _HashAlg
    SHA512: _HashAlg
    def __init__(self, hash_alg: _HashAlg) -> None: ...



# pyre-unsafe

from typing import Any

from django.core.files.base import File

class UploadedFile(File):
    size: Any
    content_type: Any
    charset: Any
    content_type_extra: Any
    name: str
    def __init__(
        self,
        file=None,
        name=None,
        content_type=None,
        size=None,
        charset=None,
        content_type_extra=None,
    ) -> None: ...
    def read(self) -> bytes: ...

class InMemoryUploadedFile(UploadedFile):
    field_name: Any
    def __init__(
        self,
        file,
        field_name,
        name,
        content_type,
        size,
        charset,
        content_type_extra=None,
    ) -> None: ...

class SimpleUploadedFile(File):
    def __init__(self, name, content, content_type="text/plain") -> None: ...



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import glob
import hashlib
import json
import logging
import multiprocessing
import os
import shutil
import site
import subprocess
import sys
from logging import Logger
from pathlib import Path
from typing import Dict, List, Optional, Union

from .exceptions import EnvironmentException
from .filesystem import assert_readable_directory, expand_relative_path, find_root
from .find_directories import (
    BINARY_NAME,
    CONFIGURATION_FILE,
    LOCAL_CONFIGURATION_FILE,
    find_typeshed,
)


LOG: Logger = logging.getLogger(__name__)


class InvalidConfiguration(Exception):
    pass


class SearchPathElement:
    def __init__(self, root: str, subdirectory: Optional[str] = None) -> None:
        self.root = os.path.expanduser(root)
        self.subdirectory = subdirectory

    @staticmethod
    def expand(
        path: Union[Dict[str, str], str], base: Optional[str] = None
    ) -> "SearchPathElement":
        if isinstance(path, str):
            if base:
                path = expand_relative_path(base, path)
            return SearchPathElement(path)
        else:
            if "root" in path and "subdirectory" in path:
                root = path["root"]
                subdirectory = path["subdirectory"]
                if base:
                    root = expand_relative_path(base, root)
                return SearchPathElement(root, subdirectory)
            elif "site-package" in path:
                site_root = site.getsitepackages()
                subdirectory = path["site-package"]

                found_element = None
                for root in site_root:
                    site_package_element = SearchPathElement(root, subdirectory)
                    if os.path.isdir(site_package_element.path()):
                        found_element = site_package_element
                if found_element is None:
                    raise InvalidConfiguration(
                        "Cannot find site package '{}'".format(subdirectory)
                    )
                return found_element
            else:
                raise InvalidConfiguration(
                    "Search path elements must have `root` and `subdirectory` "
                    "specified."
                )

    def path(self) -> str:
        subdirectory = self.subdirectory
        if subdirectory is not None:
            return os.path.join(self.root, subdirectory)
        else:
            return self.root

    def command_line_argument(self) -> str:
        subdirectory = self.subdirectory
        if subdirectory is not None:
            return self.root + "$" + subdirectory
        else:
            return self.root

    def __eq__(self, other: str) -> bool:
        # We support this for testing.
        if isinstance(other, str):
            return self.path() == other
        else:
            return self.root == other.root and self.subdirectory == other.subdirectory


class _ConfigurationFile:
    def __init__(self, file) -> None:
        self._deprecated = {"do_not_check": "ignore_all_errors"}
        contents = file.read()
        self.file_hash: str = hashlib.sha1(contents.encode("utf-8")).hexdigest()
        self._configuration = json.loads(contents)

    def consume(
        self,
        key,
        default=None,
        current=None,
        print_on_success=False,
        raise_on_override=False,
    ):
        """
        Consume a key from the configuration. When a key is consumed, it
        is removed from the configuration.

        If not found, the default is returned. If the current value is not
        None, it will be returned instead, but the key will still be
        considered consumed.
        """

        value = self._configuration.pop(key, default)
        if raise_on_override and current and value:
            raise EnvironmentException(
                "Configuration file may not override `{}` field.".format(key)
            )
        if current:
            return current
        if value and print_on_success:
            LOG.debug("Found %s: `%s`", key, ", ".join(value))
        if value and key in self._deprecated:
            LOG.warning(
                "Configuration file uses deprecated item `%s`: "
                "please migrate to its replacement `%s`",
                key,
                self._deprecated[key],
            )
        return value

    def unused_keys(self):
        """
        Return all keys not consumed yet. Some keys are explicitly whitelisted.
        """
        return self._configuration.keys() - {
            "buck_mode",
            "differential",
            "stable_client",
            "unstable_client",
            "saved_state",
            "taint_models_path",
        }


class Configuration:
    disabled: bool = False

    def __init__(
        self,
        project_root: str,
        local_root: Optional[str] = None,
        search_path: Optional[List[str]] = None,
        binary: Optional[str] = None,
        typeshed: Optional[str] = None,
        buck_builder_binary: Optional[str] = None,
        excludes: Optional[List[str]] = None,
        formatter: Optional[str] = None,
        logger: Optional[str] = None,
        log_directory: Optional[str] = None,
    ) -> None:
        self.source_directories = []
        self.targets = []
        self.logger = logger
        self.formatter = formatter
        self.ignore_all_errors = []
        self.number_of_workers: int = 0
        self.project_root: str = project_root
        self._local_root: Optional[str] = local_root
        self.taint_models_path: List[str] = []
        self.file_hash: Optional[str] = None
        self.extensions: List[str] = []
        self._log_directory: Optional[str] = log_directory

        self._version_hash: Optional[str] = None
        self._binary: Optional[str] = None
        self._typeshed: Optional[str] = None
        self._buck_builder_binary: Optional[str] = None
        self.strict: bool = False
        self._use_buck_builder: Optional[bool] = None
        self.ignore_infer: List[str] = []

        self._search_path: List[SearchPathElement] = []
        if search_path:
            search_path_elements = [
                SearchPathElement.expand(path) for path in search_path
            ]
            self._search_path.extend(search_path_elements)
        # We will extend the search path further, with the config file
        # items, inside _read().

        if binary:
            self._binary = binary

        if typeshed:
            self._typeshed = typeshed

        if buck_builder_binary:
            self._buck_builder_binary = buck_builder_binary

        self.excludes: List[str] = []
        if excludes:
            self.excludes.extend(excludes)

        if local_root:
            self._check_read_local_configuration(local_root)
        if log_directory:
            self.ignore_all_errors.append(log_directory)

        self.autocomplete = False

        self.other_critical_files: List[str] = []

        project_configuration = os.path.join(project_root, CONFIGURATION_FILE)

        # Order matters. The values will only be updated if a field is None.
        self._read(project_configuration)
        self._override_version_hash()
        self._resolve_versioned_paths()
        self._apply_defaults()
        self._validate()

    def _validate(self) -> None:
        try:

            def is_list_of_strings(list):
                if len(list) == 0:
                    return True
                return not isinstance(list, str) and all(
                    isinstance(element, str) for element in list
                )

            if not is_list_of_strings(
                self.source_directories
            ) or not is_list_of_strings(self.targets):
                raise InvalidConfiguration(
                    "`target` and `source_directories` fields must be lists of "
                    "strings."
                )

            if not is_list_of_strings(self.ignore_all_errors):
                raise InvalidConfiguration(
                    "`ignore_all_errors` field must be a list of strings."
                )

            if not is_list_of_strings(self.ignore_infer):
                raise InvalidConfiguration(
                    "`ignore_infer` field must be a list of strings."
                )

            if not is_list_of_strings(self.extensions):
                raise InvalidConfiguration(
                    "`extensions` field must be a list of strings."
                )
            if not all(
                extension.startswith(".") or not extension
                for extension in self.extensions
            ):
                raise InvalidConfiguration(
                    "`extensions` must only contain strings formatted as `.EXT`"
                )

            if not os.path.exists(self.binary):
                raise InvalidConfiguration(
                    "Binary at `{}` does not exist.".format(self.binary)
                )

            if self.number_of_workers < 1:
                raise InvalidConfiguration("Number of workers must be greater than 0.")

            # Validate typeshed path and sub-elements.
            assert_readable_directory(self.typeshed)

            # A courtesy warning since we have changed default behaviour.
            if self._typeshed_has_obsolete_value():
                LOG.warning(
                    "It appears that `{}` points at a `stdlib` directory. "
                    "Please note that the `typeshed` configuration must point at "
                    "the root of the `typeshed` directory.".format(self.typeshed)
                )

            expanded_ignore_paths = []
            for path in self.ignore_all_errors:
                expanded = glob.glob(path)
                if not expanded:
                    expanded_ignore_paths.append(path)
                else:
                    expanded_ignore_paths += expanded
            self.ignore_all_errors = expanded_ignore_paths

            non_existent_ignore_paths = [
                path for path in self.ignore_all_errors if not os.path.exists(path)
            ]
            if non_existent_ignore_paths:
                LOG.warning(
                    "Nonexistent paths passed in to `ignore_all_errors`: `{}`".format(
                        non_existent_ignore_paths
                    )
                )
                self.ignore_all_errors = [
                    path
                    for path in self.ignore_all_errors
                    if path not in non_existent_ignore_paths
                ]

            non_existent_infer_paths = [
                path for path in self.ignore_infer if not os.path.exists(path)
            ]
            if non_existent_infer_paths:
                LOG.warning(
                    "Nonexistent paths passed in to `ignore_infer`: `{}`".format(
                        non_existent_infer_paths
                    )
                )
                self.ignore_infer = [
                    path
                    for path in self.ignore_infer
                    if path not in non_existent_infer_paths
                ]

            for typeshed_subdirectory_name in ["stdlib", "third_party"]:
                typeshed_subdirectory = os.path.join(
                    self.typeshed, typeshed_subdirectory_name
                )
                assert_readable_directory(typeshed_subdirectory)
                for typeshed_version_directory_name in os.listdir(
                    typeshed_subdirectory
                ):
                    if not typeshed_version_directory_name[0].isdigit():
                        raise InvalidConfiguration(
                            "Directories inside `typeshed` must only contain "
                            "second-level subdirectories starting with "
                            "a version number."
                        )
                    typeshed_version_directory = os.path.join(
                        typeshed_subdirectory, typeshed_version_directory_name
                    )
                    assert_readable_directory(typeshed_version_directory)

            # Validate elements of the search path.
            for element in self._search_path:
                assert_readable_directory(
                    element.path(),
                    error_message_prefix="Invalid search_path in configuration: ",
                )

            if not is_list_of_strings(self.other_critical_files):
                raise InvalidConfiguration(
                    "`critical_files` field must be a list of strings."
                )
        except InvalidConfiguration as error:
            raise EnvironmentException("Invalid configuration: {}".format(str(error)))

    @property
    def version_hash(self) -> str:
        return self._version_hash or "unversioned"

    @property
    def binary(self) -> str:
        binary = self._binary
        if not binary:
            raise InvalidConfiguration("Configuration was not validated")
        return binary

    @property
    def typeshed(self) -> str:
        typeshed = self._typeshed
        if not typeshed:
            raise InvalidConfiguration("Configuration invalid: no typeshed specified")
        return typeshed

    @property
    def buck_builder_binary(self) -> Optional[str]:
        return self._buck_builder_binary

    @property
    def use_buck_builder(self) -> bool:
        return self._use_buck_builder or False

    @property
    def search_path(self) -> List[str]:
        if not self._search_path:
            return []
        return [element.command_line_argument() for element in self._search_path]

    @property
    def local_root(self) -> Optional[str]:
        return self._local_root

    @property
    def log_directory(self) -> str:
        log_directory = self._log_directory
        if not log_directory:
            raise InvalidConfiguration("Configuration was not validated")
        return log_directory

    def get_binary_version(self) -> Optional[str]:
        status = subprocess.run(
            [self.binary, "-version"], stdout=subprocess.PIPE, universal_newlines=True
        )
        if status.returncode == 0:
            return status.stdout.strip()
        else:
            return None

    def _check_nested_configurations(self, local_root: str) -> None:
        # TODO(T67874463): Handle sanity checks against project configurations
        parent_local_root = find_root(
            os.path.dirname(local_root.rstrip("/")), LOCAL_CONFIGURATION_FILE
        )
        if not parent_local_root:
            return

        parent_error = None
        excluded_from_parent = False
        local_root: Path = Path(local_root).resolve()
        try:
            parent_local_configuration = Configuration(
                project_root=self.project_root,
                local_root=parent_local_root,
                search_path=[search_path.root for search_path in self._search_path],
                binary=self._binary,
                typeshed=self._typeshed,
                buck_builder_binary=self._buck_builder_binary,
                excludes=self.excludes,
                logger=self.logger,
                formatter=self.formatter,
                log_directory=self._log_directory,
            )
            paths_to_ignore = list(local_root.parents) + [local_root]
            for ignore_element in parent_local_configuration.ignore_all_errors:
                if Path(ignore_element).resolve() in paths_to_ignore:
                    excluded_from_parent = True
        except EnvironmentException as error:
            parent_error = error

        if not excluded_from_parent:
            relative_path = str(local_root.relative_to(parent_local_root))
            error_message = (
                "Local configuration is nested under another local configuration at "
                f"`{parent_local_root}`.\nPlease add `{relative_path}` to the "
                "`ignore_all_errors` field of the parent, or combine the sources "
                "into a single configuration, or split the parent configuration to "
                "avoid inconsistent errors."
            )
            raise EnvironmentException(error_message)
        elif parent_error:
            raise EnvironmentException(parent_error)

    def _check_read_local_configuration(self, path: str) -> None:
        if not os.path.exists(path):
            raise EnvironmentException(
                "Local configuration path `{}` does not exist.".format(path)
            )

        local_configuration = os.path.join(path, LOCAL_CONFIGURATION_FILE)
        if not os.path.exists(local_configuration):
            raise EnvironmentException(
                "Local configuration directory `{}` does not contain "
                "a `{}` file.".format(path, LOCAL_CONFIGURATION_FILE)
            )

        self._check_nested_configurations(path)
        self._read(local_configuration)
        if not self.source_directories and not self.targets:
            raise EnvironmentException(
                "Local configuration `{}` does not specify any sources to type check.".format(
                    local_configuration
                )
            )

    def _read(self, path: str) -> None:
        try:
            with open(path) as file:
                LOG.debug("Reading configuration `%s`...", path)
                configuration_directory = os.path.dirname(os.path.realpath(path))
                configuration = _ConfigurationFile(file)

                source_directories = configuration.consume(
                    "source_directories",
                    default=[],
                    current=self.source_directories,
                    print_on_success=True,
                    raise_on_override=True,
                )
                self.source_directories = [
                    expand_relative_path(configuration_directory, directory)
                    for directory in source_directories
                ]

                self.targets = configuration.consume(
                    "targets",
                    default=[],
                    current=self.targets,
                    print_on_success=True,
                    raise_on_override=True,
                )

                if configuration.consume("disabled", default=False):
                    self.disabled = True

                logger = configuration.consume("logger", current=self.logger)
                if logger:
                    self.logger = expand_relative_path(configuration_directory, logger)

                self.formatter = configuration.consume(
                    "formatter", current=self.formatter
                )

                self.strict = configuration.consume("strict", default=self.strict)

                ignore_all_errors = configuration.consume(
                    "ignore_all_errors", default=[]
                )
                self.ignore_all_errors.extend(
                    [
                        expand_relative_path(root=configuration_directory, path=path)
                        for path in ignore_all_errors
                    ]
                )

                ignore_infer = configuration.consume("ignore_infer", default=[])
                self.ignore_infer.extend(
                    [
                        expand_relative_path(root=configuration_directory, path=path)
                        for path in ignore_infer
                    ]
                )

                self.number_of_workers = int(
                    configuration.consume(
                        "workers", default=0, current=self.number_of_workers
                    )
                )

                binary = configuration.consume("binary", current=self._binary)
                assert binary is None or isinstance(binary, str)
                if binary is not None:
                    binary = expand_relative_path(configuration_directory, binary)
                self._binary = binary

                buck_builder_binary = configuration.consume(
                    "buck_builder_binary", current=self._buck_builder_binary
                )
                if buck_builder_binary is not None:
                    self._buck_builder_binary = expand_relative_path(
                        root=configuration_directory, path=buck_builder_binary
                    )

                additional_search_path = configuration.consume(
                    "search_path", default=[]
                )

                if isinstance(additional_search_path, list):
                    self._search_path.extend(
                        [
                            SearchPathElement.expand(path, base=configuration_directory)
                            for path in additional_search_path
                        ]
                    )
                else:
                    self._search_path.append(
                        SearchPathElement(
                            expand_relative_path(
                                configuration_directory, additional_search_path
                            )
                        )
                    )

                version_hash = configuration.consume(
                    "version", current=self._version_hash
                )
                assert version_hash is None or isinstance(version_hash, str)
                self._version_hash = version_hash

                typeshed = configuration.consume("typeshed", current=self._typeshed)
                assert typeshed is None or isinstance(typeshed, str)
                if typeshed is not None:
                    typeshed = expand_relative_path(configuration_directory, typeshed)
                self._typeshed = typeshed

                taint_models_path = configuration.consume("taint_models_path")
                assert (
                    taint_models_path is None
                    or isinstance(taint_models_path, str)
                    or isinstance(taint_models_path, list)
                )
                if isinstance(taint_models_path, str):

                    self.taint_models_path.append(
                        expand_relative_path(configuration_directory, taint_models_path)
                    )
                elif isinstance(taint_models_path, list):
                    self.taint_models_path.extend(
                        [
                            expand_relative_path(configuration_directory, path)
                            for path in taint_models_path
                        ]
                    )

                excludes = configuration.consume("exclude", default=[])
                if isinstance(excludes, list):
                    self.excludes.extend(
                        [
                            expand_relative_path(configuration_directory, path)
                            for path in excludes
                        ]
                    )
                else:
                    self.excludes.append(
                        expand_relative_path(configuration_directory, excludes)
                    )

                extensions = configuration.consume("extensions", default=[])
                self.extensions.extend(extensions)

                # We rely on the configuration SHA1 to make
                if configuration.consume("saved_state"):
                    self.file_hash = configuration.file_hash

                use_buck_builder = configuration.consume("use_buck_builder")
                if self._use_buck_builder is None:
                    self._use_buck_builder = use_buck_builder

                self.autocomplete = configuration.consume("autocomplete", default=False)

                critical_files = configuration.consume("critical_files", default=[])
                self.other_critical_files = [
                    expand_relative_path(root=configuration_directory, path=path)
                    for path in critical_files
                ]

                # Warn on deprecated fields.
                for deprecated_field in configuration._deprecated.keys():
                    configuration.consume(deprecated_field)

                # This block should be at the bottom to be effective.
                unused_keys = configuration.unused_keys()
                if unused_keys:
                    LOG.warning(
                        "Some configuration items were not recognized in "
                        "`{}`: {}".format(path, ", ".join(unused_keys))
                    )
        except IOError:
            # To avoid TOCTTOU bugs, handle IOErrors here silently.
            # We error elsewhere if there weren't enough parameters passed into pyre.
            pass
        except json.JSONDecodeError as error:
            raise EnvironmentException(
                "Configuration file at `{}` is invalid: {}.".format(path, str(error))
            )

    def _resolve_versioned_paths(self) -> None:
        version_hash = self.version_hash
        if not version_hash:
            return

        binary = self._binary
        if binary:
            self._binary = binary.replace("%V", version_hash)
        typeshed = self._typeshed
        if typeshed:
            self._typeshed = typeshed.replace("%V", version_hash)

    def _override_version_hash(self) -> None:
        overriding_version_hash = os.getenv("PYRE_VERSION_HASH")
        if overriding_version_hash:
            self._version_hash = overriding_version_hash
            LOG.warning("Version hash overridden with `%s`", self._version_hash)

    def _apply_defaults(self) -> None:
        overriding_binary = os.getenv("PYRE_BINARY")
        if overriding_binary:
            self._binary = overriding_binary
            LOG.warning("Binary overridden with `%s`", self._binary)
        if not self._binary:
            LOG.info(
                "No binary specified, looking for `{}` in PATH".format(BINARY_NAME)
            )
            self._binary = shutil.which(BINARY_NAME)
            if not self._binary:
                binary_candidate = os.path.join(
                    os.path.dirname(sys.argv[0]), BINARY_NAME
                )
                self._binary = shutil.which(binary_candidate)
            if not self._binary:
                LOG.warning("Could not find `{}` in PATH".format(BINARY_NAME))
            else:
                LOG.info("Found: `%s`", self._binary)

        if self.number_of_workers == 0:
            try:
                self.number_of_workers = max(multiprocessing.cpu_count() - 4, 1)
            except NotImplementedError:
                self.number_of_workers = 4

        if not self._typeshed:
            LOG.info("No typeshed specified, looking for it")
            self._typeshed = find_typeshed()
            if not self._typeshed:
                LOG.warning("Could not find a suitable typeshed")
            else:
                LOG.info("Found: `%s`", self._typeshed)

        if not self._log_directory:
            # TODO(T56191177): We should not start up a server at all if no configurations
            # exist. Currently, we treat the cwd as the project root if no configurations
            # exist. Instead, we should default logging to `tmp/.pyre` in the initial
            # find_log_directory as well.
            self._log_directory = "/tmp/.pyre"

    def _typeshed_has_obsolete_value(self) -> bool:
        (head, tail) = os.path.split(self.typeshed)
        if tail == "stdlib":
            return True
        if tail != "":
            return False
        # If `path` ends in a slash, tail will be empty.
        (head, tail) = os.path.split(head)
        return tail == "stdlib"



# Stubs for collections.abc (introduced from Python 3.3)
#
# https://docs.python.org/3.3/whatsnew/3.3.html#collections
import sys

from . import (
    AsyncIterable as AsyncIterable,
    AsyncIterator as AsyncIterator,
    Awaitable as Awaitable,
    ByteString as ByteString,
    Callable as Callable,
    Container as Container,
    Coroutine as Coroutine,
    Generator as Generator,
    Hashable as Hashable,
    ItemsView as ItemsView,
    Iterable as Iterable,
    Iterator as Iterator,
    KeysView as KeysView,
    Mapping as Mapping,
    MappingView as MappingView,
    MutableMapping as MutableMapping,
    MutableSequence as MutableSequence,
    MutableSet as MutableSet,
    Sequence as Sequence,
    Set as Set,
    Sized as Sized,
    ValuesView as ValuesView,
)

if sys.version_info >= (3, 6):
    from . import AsyncGenerator as AsyncGenerator, Collection as Collection, Reversible as Reversible



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import logging
import sys
import traceback
from logging import Logger

from ...client.commands import ExitCode
from . import UserError
from .ast import UnstableAST
from .commands.codemods import (
    MissingGlobalAnnotations,
    MissingOverrideReturnAnnotations,
)
from .commands.consolidate_nested_configurations import ConsolidateNestedConfigurations
from .commands.expand_target_coverage import ExpandTargetCoverage
from .commands.fixme import Fixme
from .commands.fixme_all import FixmeAll
from .commands.fixme_single import FixmeSingle
from .commands.fixme_targets import FixmeTargets
from .commands.global_version_update import GlobalVersionUpdate
from .commands.strict_default import StrictDefault
from .commands.targets_to_configuration import TargetsToConfiguration
from .repository import Repository


LOG: Logger = logging.getLogger(__name__)


def run(repository: Repository) -> None:
    parser = argparse.ArgumentParser(fromfile_prefix_chars="@")
    parser.add_argument("--verbose", action="store_true", help="Enable verbose logging")

    commands = parser.add_subparsers()

    # Subcommands: Codemods
    missing_overridden_return_annotations = commands.add_parser(
        "missing-overridden-return-annotations",
        help="Add annotations according to errors inputted through stdin.",
    )
    MissingOverrideReturnAnnotations.add_arguments(
        missing_overridden_return_annotations
    )

    missing_global_annotations = commands.add_parser(
        "missing-global-annotations",
        help="Add annotations according to errors inputted through stdin.",
    )
    MissingGlobalAnnotations.add_arguments(missing_global_annotations)

    # Subcommand: Change default pyre mode to strict and adjust module headers.
    strict_default = commands.add_parser("strict-default")
    StrictDefault.add_arguments(strict_default)

    # Subcommand: Set global configuration to given hash, and add version override
    # to all local configurations to run previous version.
    update_global_version = commands.add_parser("update-global-version")
    GlobalVersionUpdate.add_arguments(update_global_version)

    # Subcommand: Fixme all errors inputted through stdin.
    fixme = commands.add_parser("fixme")
    Fixme.add_arguments(fixme)

    # Subcommand: Fixme all errors for a single project.
    fixme_single = commands.add_parser("fixme-single")
    FixmeSingle.add_arguments(fixme_single)

    # Subcommand: Fixme all errors in all projects with local configurations.
    fixme_all = commands.add_parser("fixme-all")
    FixmeAll.add_arguments(fixme_all)

    # Subcommand: Fixme all errors in targets running type checking
    fixme_targets = commands.add_parser("fixme-targets")
    FixmeTargets.add_arguments(fixme_targets)

    # Subcommand: Remove targets integration and replace with configuration
    targets_to_configuration = commands.add_parser("targets-to-configuration")
    TargetsToConfiguration.add_arguments(targets_to_configuration)

    # Subcommand: Expand target coverage in configuration up to given error limit
    expand_target_coverage = commands.add_parser("expand-target-coverage")
    ExpandTargetCoverage.add_arguments(expand_target_coverage)

    # Subcommand: Consolidate nested local configurations
    consolidate_nested_configurations = commands.add_parser("consolidate-nested")
    ConsolidateNestedConfigurations.add_arguments(consolidate_nested_configurations)

    # Initialize default values.
    arguments = parser.parse_args()
    if not hasattr(arguments, "command"):
        # Reparsing with `fixme` as default subcommand.
        arguments = parser.parse_args(sys.argv[1:] + ["fixme"])

    logging.basicConfig(
        format="%(asctime)s %(levelname)s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=logging.DEBUG if arguments.verbose else logging.INFO,
    )

    try:
        exit_code = ExitCode.SUCCESS
        arguments.command(arguments, repository).run()
    except UnstableAST as error:
        LOG.error(str(error))
        exit_code = ExitCode.FOUND_ERRORS
    except UserError as error:
        LOG.error(str(error))
        exit_code = ExitCode.FAILURE
    except Exception as error:
        LOG.error(str(error))
        LOG.debug(traceback.format_exc())
        exit_code = ExitCode.FAILURE

    sys.exit(exit_code)


def main() -> None:
    run(Repository())


if __name__ == "__main__":
    main()



# Stubs for distutils.extension

import sys
from typing import List, Optional, Tuple

class Extension:
    if sys.version_info >= (3,):
        def __init__(
            self,
            name: str,
            sources: List[str],
            include_dirs: List[str] = ...,
            define_macros: List[Tuple[str, Optional[str]]] = ...,
            undef_macros: List[str] = ...,
            library_dirs: List[str] = ...,
            libraries: List[str] = ...,
            runtime_library_dirs: List[str] = ...,
            extra_objects: List[str] = ...,
            extra_compile_args: List[str] = ...,
            extra_link_args: List[str] = ...,
            export_symbols: List[str] = ...,
            swig_opts: Optional[str] = ...,  # undocumented
            depends: List[str] = ...,
            language: str = ...,
            optional: bool = ...,
        ) -> None: ...
    else:
        def __init__(
            self,
            name: str,
            sources: List[str],
            include_dirs: List[str] = ...,
            define_macros: List[Tuple[str, Optional[str]]] = ...,
            undef_macros: List[str] = ...,
            library_dirs: List[str] = ...,
            libraries: List[str] = ...,
            runtime_library_dirs: List[str] = ...,
            extra_objects: List[str] = ...,
            extra_compile_args: List[str] = ...,
            extra_link_args: List[str] = ...,
            export_symbols: List[str] = ...,
            swig_opts: Optional[str] = ...,  # undocumented
            depends: List[str] = ...,
            language: str = ...,
        ) -> None: ...



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from collections import Counter
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple

from .models import SharedTextKind, TraceFrame, TraceFrameAnnotation, TraceKind
from .trace_graph import TraceGraph


class TrimmedTraceGraph(TraceGraph):
    """Represents a trimmed graph that is constructed from a bigger TraceGraph
    based on issues that have traces involving a set of affected files or
    directories.
    """

    def __init__(
        self, affected_files: List[str], affected_issues_only: bool = False
    ) -> None:
        """Creates an empty TrimmedTraceGraph.
        """
        super().__init__()
        self._affected_files = affected_files
        self._affected_issues_only = affected_issues_only
        self._visited_trace_frame_ids: Set[int] = set()

    def populate_from_trace_graph(self, graph: TraceGraph) -> None:
        """Populates this graph from the given one based on affected_files
        """
        # Track which trace frames have been visited as we populate the full
        # traces of the graph.
        self._visited_trace_frame_ids: Set[int] = set()

        self._populate_affected_issues(graph)

        if not self._affected_issues_only:
            # Finds issues from the conditions and saves them.
            # Also saves traces that have been trimmed to the affected
            # conditions.
            self._populate_issues_from_affected_trace_frames(graph)

            # Traces populated above may be missing all traces because
            # _populate_issues_from_affected_trace_frames only populates
            # traces that reach the affected conditions in one direction. We
            # may need to populate traces in other directions too.
            #
            # For example:
            #
            # Issue_x reaches affected_file_x via postcondition_x (forward
            # trace, i.e. trace leading to source). None of its backward
            # traces (leading to sinks) reach the affected files.
            #
            # _populate_issues_from_affected_trace_frames would have copied its
            # forward traces and trimmed it to those reaching postcondition_x.
            # We cannot blindly populate all forward traces in this case as
            # branches not leading to postcondition_x are unnecessary.
            #
            # However, in this specific example, all backward traces are needed
            # to give a complete picture of which sinks the issue reaches.
            # The following ensures that.
            for instance_id in self._issue_instances.keys():
                first_hop_ids = self._issue_instance_trace_frame_assoc[instance_id]
                fwd_trace_ids = {
                    tf_id
                    for tf_id in first_hop_ids
                    if self._trace_frames[tf_id].kind == TraceKind.POSTCONDITION
                }
                bwd_trace_ids = {
                    tf_id
                    for tf_id in first_hop_ids
                    if self._trace_frames[tf_id].kind == TraceKind.PRECONDITION
                }

                if len(fwd_trace_ids) == 0:
                    self._populate_issue_trace(
                        graph,
                        instance_id,
                        # pyre-fixme[6]: Expected `Optional[TraceKind]` for 3rd
                        #  param but got `(cls: TraceKind) -> Any`.
                        TraceKind.POSTCONDITION,
                    )

                if len(bwd_trace_ids) == 0:
                    self._populate_issue_trace(
                        graph,
                        instance_id,
                        # pyre-fixme[6]: Expected `Optional[TraceKind]` for 3rd
                        #  param but got `(cls: TraceKind) -> Any`.
                        TraceKind.PRECONDITION,
                    )

        self._recompute_instance_properties()

    def _recompute_instance_properties(self):
        """Some properties of issue instances will be affected after trimming
        such as min trace length to leaves. This should be called after the
        trimming to re-compute these values.
        """
        callables_histo = Counter(
            inst.callable_id.local_id for inst in self._issue_instances.values()
        )

        for inst in self._issue_instances.values():
            inst.min_trace_length_to_sources = self._get_min_depth_to_sources(
                inst.id.local_id
            )
            inst.min_trace_length_to_sinks = self._get_min_depth_to_sinks(
                inst.id.local_id
            )
            inst.callable_count = callables_histo[inst.callable_id.local_id]

    def _get_min_depth_to_sources(self, instance_id: int) -> int:
        """Returns shortest depth to source from the issue instance. Instances
        have a pre-computed min_trace_length_to_source, but this can change
        after traces get trimmed from the graph. This re-computes it and
        returns the min.
        """
        source_ids = {
            source.id.local_id
            for source in (
                # pyre-fixme[6]: Expected `SharedTextKind` for 2nd param but got
                #  `(cls: SharedTextKind) -> Any`.
                self.get_issue_instance_shared_texts(instance_id, SharedTextKind.SOURCE)
            )
        }
        first_hop_tf_ids = {
            tf_id
            for tf_id in self._issue_instance_trace_frame_assoc[instance_id]
            if self.get_trace_frame_from_id(tf_id).kind == TraceKind.POSTCONDITION
        }
        return self._get_min_depth(first_hop_tf_ids, source_ids)

    def _get_min_depth_to_sinks(self, instance_id: int) -> int:
        """See get_min_depths_to_sources."""
        sink_ids = {
            sink.id.local_id
            for sink in (
                # pyre-fixme[6]: Expected `SharedTextKind` for 2nd param but got
                #  `(cls: SharedTextKind) -> Any`.
                self.get_issue_instance_shared_texts(instance_id, SharedTextKind.SINK)
            )
        }
        first_hop_tf_ids = {
            tf_id
            for tf_id in self._issue_instance_trace_frame_assoc[instance_id]
            if self.get_trace_frame_from_id(tf_id).kind == TraceKind.PRECONDITION
        }
        return self._get_min_depth(first_hop_tf_ids, sink_ids)

    def _get_min_depth(self, first_hop_tf_ids: Set[int], leaf_ids: Set[int]) -> int:
        min_depth = None
        for tf_id in first_hop_tf_ids:
            leaf_depths = self._trace_frame_leaf_assoc[tf_id]
            for (leaf_id, depth) in leaf_depths:
                if leaf_id in leaf_ids and (min_depth is None or depth < min_depth):
                    min_depth = depth
        if min_depth is not None:
            return min_depth
        return 0

    def _populate_affected_issues(self, graph: TraceGraph) -> None:
        """Populates the trimmed graph with issues whose locations are in
        affected_files based on data in the input graph. Since these issues
        exist in the affected files, all traces are copied as well.
        """
        affected_instance_ids = [
            instance.id.local_id
            for instance in graph._issue_instances.values()
            if self._is_filename_prefixed_with(
                graph.get_text(instance.filename_id), self._affected_files
            )
        ]

        for instance_id in affected_instance_ids:
            if instance_id in self._issue_instances:
                continue
            self._populate_issue_and_traces(graph, instance_id)

    def _get_sink_names(self, graph: TraceGraph, instance_id: int) -> Set[str]:
        # pyre-fixme[9]: kind has type `SharedTextKind`; used as `(cls:
        #  SharedTextKind) -> Any`.
        kind: SharedTextKind = SharedTextKind.SINK
        sinks = graph.get_issue_instance_shared_texts(instance_id, kind)
        return {sink.contents for sink in sinks}

    def _get_source_names(self, graph: TraceGraph, instance_id: int) -> Set[str]:
        # pyre-fixme[9]: kind has type `SharedTextKind`; used as `(cls:
        #  SharedTextKind) -> Any`.
        kind: SharedTextKind = SharedTextKind.SOURCE
        sources = graph.get_issue_instance_shared_texts(instance_id, kind)
        return {source.contents for source in sources}

    def _get_leaf_names(self, graph: TraceGraph, instance_id: int) -> Set[str]:
        return self._get_source_names(graph, instance_id).union(
            self._get_sink_names(graph, instance_id)
        )

    def _get_leaf_names_from_pairs(
        self, graph: TraceGraph, leaf_pairs: Set[Tuple[int, Any]]
    ) -> Set[str]:
        return {
            leaf.contents
            for leaf in (graph._shared_texts[leaf_id] for (leaf_id, _) in leaf_pairs)
        }

    def _populate_issues_from_affected_trace_frames(self, graph: TraceGraph) -> None:
        """TraceFrames found in affected_files should be reachable via some
        issue instance. Follow traces of graph to find them and
        populate this TrimmedGraph with it.
        """

        initial_trace_frames = [
            trace_frame
            for trace_frame in graph._trace_frames.values()
            if self._is_filename_prefixed_with(
                graph.get_text(trace_frame.filename_id), self._affected_files
            )
        ]

        self._populate_issues_from_affected_conditions(
            initial_trace_frames,
            graph,
            lambda trace_frame_id: (
                graph._trace_frame_issue_instance_assoc[trace_frame_id]
            ),
            lambda trace_frame: (
                [
                    graph._trace_frames[trace_frame_id]
                    for trace_frame_id in graph._trace_frames_rev_map[trace_frame.kind][
                        (trace_frame.caller_id.local_id, trace_frame.caller_port)
                    ]
                ]
            ),
            lambda instance_id: (self._get_leaf_names(graph, instance_id)),
            lambda trace_frame_id: (
                self._get_leaf_names_from_pairs(
                    graph, graph._trace_frame_leaf_assoc[trace_frame_id]
                )
            ),
            lambda instance, trace_frame: (
                self.add_issue_instance_trace_frame_assoc(instance, trace_frame)
            ),
            lambda trace_frame_id: (
                self._add_trace_frame(graph, graph._trace_frames[trace_frame_id])
            ),
            lambda initial_trace_frame_ids: (
                self._populate_trace(graph, initial_trace_frame_ids)
            ),
        )

    def _populate_issues_from_affected_conditions(
        self,
        initial_conditions,
        graph: TraceGraph,
        get_issue_instances_from_condition_id,
        get_condition_parent,
        get_instance_leaves,
        get_condition_leaves,
        add_instance_condition_assoc,
        add_condition,
        add_traces,
    ) -> None:
        """Helper for populating reachable issue instances from the initial
        pre/postconditions. Also populates conditions/traces reachable from
        these instances. Traces are populated only in the direction this is
        called from: i.e. if initial_conditions are preconditions, only the
        backward trace is populated.

        Params:
        initial_conditions: The initial collection of pre/postconditions to
        start searching for issues from.

        graph: The trace graph to search for issues. Nodes/edges in this graph
        will be copied over to the local state

        get_issue_instances_from_condition_id: Function that returns all
        issue instances associated with a given a pre/postcondition id.

        get_condition_parent: Function that returns the parent condition of
        a given pre/postcondition. Given a pre/postcondition, p, its parent p',
        is the pre/postcondition that calls it, i.e. p.caller = p'.callee

        get_instance_leaves: Function that returns a collection of leaf names
        associated with the given issue instance ID.

        get_condition_leaves: Function that returns a collection of leaf names
        associated with the given (pre/post)condition ID.

        add_instance_condition_assoc: Function that takes in the issue
        instance and condition and adds the assoc between them.

        add_condition: Function that adds a condition to the graph given its
        id. This must add all the condition's assocs with the leaves because
        we don't filter out any condition-leaf assocs.

        add_traces: Function that takes a list of initial conditions and
        adds all conditions reachable from these to the graph.
        """
        visited: Dict[int, Set[str]] = {}
        que = [
            (cond, get_condition_leaves(cond.id.local_id))
            for cond in initial_conditions
        ]

        # Note that parent conditions may not transitively lead to the leaves
        # that its descendents lead to due to special-cased leaf filtering at
        # analysis time. When visiting each condition, we need to track the
        # leaves that we are visiting it from and only visit parent traces that
        # share common leaves along the path.
        while len(que) > 0:
            condition, leaves = que.pop()
            cond_id = condition.id.local_id

            if cond_id in visited:
                leaves = leaves - visited[cond_id]
                if len(leaves) == 0:
                    continue
                else:
                    visited[cond_id].update(leaves)
            else:
                visited[cond_id] = leaves

            # Found instance(s) related to the current condition. Yay.
            # This instance may have been found before, but process it again
            # anyway because we need to add the assoc with this condition.
            for instance_id in get_issue_instances_from_condition_id(cond_id):
                # Check if the leaves (sources/sinks) of the issue reach
                # the same leaves as the ones relevant to this condition.
                instance = graph._issue_instances[instance_id]
                issue_leaves = set(get_instance_leaves(instance.id.local_id))
                common_leaves = issue_leaves.intersection(leaves)
                if len(common_leaves) > 0:
                    if instance_id not in self._issue_instances:
                        self._populate_issue(graph, instance_id)
                    add_instance_condition_assoc(instance, condition)

            # Conditions that call this may have originated from other issues,
            # keep searching for parent conditions leading to this one.
            for next_cond in get_condition_parent(condition):
                next_leaves = get_condition_leaves(next_cond.id.local_id)
                common_leaves = leaves.intersection(next_leaves)
                if len(leaves) > 0:
                    que.append((next_cond, common_leaves))

        # Add traces leading out from initial_conditions, and all visited
        # conditions leading back towards the issues.
        initial_condition_ids = [
            condition.id.local_id for condition in initial_conditions
        ]
        add_traces(initial_condition_ids)
        for condition_id in visited:
            add_condition(condition_id)

    def _populate_issue_and_traces(self, graph: TraceGraph, instance_id: int) -> None:
        """ Copies an issue over from the given trace graph, including all its
        traces and assocs.
        """
        self._populate_issue(graph, instance_id)
        self._populate_issue_trace(graph, instance_id)

    def _populate_issue_trace(
        self, graph: TraceGraph, instance_id: int, kind: Optional[TraceKind] = None
    ) -> None:
        trace_frame_ids = list(graph._issue_instance_trace_frame_assoc[instance_id])
        instance = graph._issue_instances[instance_id]
        filtered_ids = []
        for trace_frame_id in trace_frame_ids:
            frame = graph._trace_frames[trace_frame_id]
            if kind is None or kind == frame.kind:
                self.add_issue_instance_trace_frame_assoc(instance, frame)
                filtered_ids.append(trace_frame_id)
        self._populate_trace(graph, filtered_ids)

    def _populate_issue(self, graph: TraceGraph, instance_id: int) -> None:
        """Adds an issue to the trace graph along with relevant information
        pertaining to the issue (e.g. instance, fix_info, sources/sinks)
        The issue is identified by its corresponding instance's ID in the input
        trace graph.
        """
        instance = graph._issue_instances[instance_id]
        issue = graph._issues[instance.issue_id.local_id]
        self._populate_shared_text(graph, instance.message_id)
        self._populate_shared_text(graph, instance.filename_id)
        self._populate_shared_text(graph, instance.callable_id)

        self.add_issue_instance(instance)
        self.add_issue(issue)

        if instance_id in graph._issue_instance_fix_info:
            issue_fix_info = graph._issue_instance_fix_info[instance_id]
            self.add_issue_instance_fix_info(instance, issue_fix_info)

        for shared_text_id in graph._issue_instance_shared_text_assoc[instance_id]:
            shared_text = graph._shared_texts[shared_text_id]
            if shared_text_id not in self._shared_texts:
                self.add_shared_text(shared_text)
            self.add_issue_instance_shared_text_assoc(instance, shared_text)

    def _populate_trace(self, graph: TraceGraph, trace_frame_ids: List[int]) -> None:
        """ Populates (from the given trace graph) the forward and backward
        traces reachable from the given traces (including input trace frames).
        Make sure to respect trace kind in successors
        """
        while len(trace_frame_ids) > 0:
            trace_frame_id = trace_frame_ids.pop()
            if trace_frame_id in self._visited_trace_frame_ids:
                continue

            trace_frame = graph._trace_frames[trace_frame_id]
            self._add_trace_frame(graph, trace_frame)
            self._visited_trace_frame_ids.add(trace_frame_id)

            trace_frame_ids.extend(
                [
                    next_frame.id.local_id
                    for next_frame in graph.get_next_trace_frames(trace_frame)
                    if next_frame.id.local_id not in self._visited_trace_frame_ids
                ]
            )

    def _add_trace_frame(self, graph: TraceGraph, trace_frame: TraceFrame) -> None:
        """ Copies the trace frame from 'graph' to this (self) graph.
        Also copies all the trace_frame-leaf assocs since we don't
        know which ones are needed until we know the issue that reaches it
        """
        trace_frame_id = trace_frame.id.local_id
        self.add_trace_frame(trace_frame)

        annotations = graph.get_condition_annotations(trace_frame_id)
        for annotation in annotations:
            self._add_trace_annotation(graph, annotation)

        self._populate_shared_text(graph, trace_frame.filename_id)
        self._populate_shared_text(graph, trace_frame.caller_id)
        self._populate_shared_text(graph, trace_frame.callee_id)
        for (leaf_id, depth) in graph._trace_frame_leaf_assoc[trace_frame_id]:
            leaf = graph._shared_texts[leaf_id]
            if leaf_id not in self._shared_texts:
                self.add_shared_text(leaf)
            self.add_trace_frame_leaf_assoc(trace_frame, leaf, depth)

    @staticmethod
    def _is_filename_prefixed_with(filename: str, prefixes: Iterable[str]) -> bool:
        return any([filename.startswith(p) for p in prefixes])

    def _populate_shared_text(self, graph, id) -> None:
        text = graph._shared_texts[id.local_id]
        if text.id.local_id not in self._shared_texts:
            self.add_shared_text(text)

    def _add_trace_annotation(
        self, graph: TraceGraph, annotation: TraceFrameAnnotation
    ) -> None:
        """Copies the annotation from 'graph' to this (self) graph.
        Also copies children TraceFrames of the annotation (if any). The
        parent TraceFrame of the annotation is NOT copied.
        """
        self.add_trace_annotation(annotation)
        children = graph.get_annotation_trace_frames(annotation.id.local_id)
        child_ids = [child.id.local_id for child in children]
        for child in children:
            self.add_trace_frame_annotation_trace_frame_assoc(annotation, child)
        self._populate_trace(graph, child_ids)



import contextvars
from typing import Callable, List, Sequence, Tuple
from typing_extensions import Literal

from . import futures

_PENDING: Literal["PENDING"]  # undocumented
_CANCELLED: Literal["CANCELLED"]  # undocumented
_FINISHED: Literal["FINISHED"]  # undocumented

def isfuture(obj: object) -> bool: ...
def _format_callbacks(cb: Sequence[Tuple[Callable[[futures.Future], None], contextvars.Context]]) -> str: ...  # undocumented
def _future_repr_info(future: futures.Future) -> List[str]: ...  # undocumented



from django.core.exceptions import ValidationError
from django.forms.fields import BooleanField, EmailField
from django.forms.forms import Form
from django.forms.models import ModelForm
from django.forms.widgets import (
    CheckboxInput,
    EmailInput,
    HiddenInput,
    NumberInput,
    TextInput,
    URLInput,
)



# pyre-unsafe

from typing import Any, Dict, Optional, Union

class BaseCookie(dict):
    def __init__(self, input: Optional[Union[str, Dict[str, str]]] = None) -> None: ...

class SimpleCookie(BaseCookie):
    def __getitem__(self, key: str) -> Any: ...
    def update(self, other: Any) -> None: ...
    def get(self, key: Any) -> Any: ...
    def load(self, vals: Dict[Any, Any]) -> None: ...



# pyre-unsafe

from typing import Any, Optional

from django.http import HttpRequest, HttpResponse
from django.http.response import HttpResponseBase

def get_token(request: HttpRequest) -> Optional[str]: ...
def rotate_token(request: HttpRequest) -> None: ...

class CsrfViewMiddleware:
    def process_view(
        self,
        request: HttpRequest,
        callback: Any,
        callback_args: Any,
        callback_kwargs: Any,
    ) -> Optional[HttpResponse]: ...
    def process_response(
        self, request: HttpRequest, response: HttpResponseBase
    ) -> HttpResponseBase: ...



import sys
from typing import Any, Callable, Iterator, Optional, Tuple

class JSONEncoder:
    item_separator: str
    key_separator: str

    skipkeys: bool
    ensure_ascii: bool
    check_circular: bool
    allow_nan: bool
    sort_keys: bool
    indent: int

    if sys.version_info >= (3, 6):
        def __init__(
            self,
            *,
            skipkeys: bool = ...,
            ensure_ascii: bool = ...,
            check_circular: bool = ...,
            allow_nan: bool = ...,
            sort_keys: bool = ...,
            indent: Optional[int] = ...,
            separators: Optional[Tuple[str, str]] = ...,
            default: Optional[Callable[..., Any]] = ...,
        ) -> None: ...
    else:
        def __init__(
            self,
            skipkeys: bool = ...,
            ensure_ascii: bool = ...,
            check_circular: bool = ...,
            allow_nan: bool = ...,
            sort_keys: bool = ...,
            indent: Optional[int] = ...,
            separators: Optional[Tuple[str, str]] = ...,
            default: Optional[Callable[..., Any]] = ...,
        ) -> None: ...
    def default(self, o: Any) -> Any: ...
    def encode(self, o: Any) -> str: ...
    def iterencode(self, o: Any, _one_shot: bool = ...) -> Iterator[str]: ...



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import sys
from typing import TextIO


def is_capable(file: TextIO = sys.stderr) -> bool:
    """
    Determine whether we are connected to a capable terminal.
    """
    if not os.isatty(file.fileno()):
        return False
    terminal = os.getenv("TERM", "dumb")
    # Hardcoded list of non-capable terminals.
    return terminal not in ["dumb", "emacs"]



# pyre-unsafe

from typing import Any

NON_FIELD_ERRORS: str

class PermissionDenied(Exception):
    pass

class ViewDoesNotExist(Exception):
    pass

class MiddlewareNotUsed(Exception):
    pass

class ValidationError(Exception):
    def __init__(self, message, code=None, params=None): ...
    error_dict: Any
    error_list: Any
    def update_error_dict(self, error_dict: Any) -> Any: ...
    message: Any
    code: Any

class SuspiciousOperation(Exception):
    pass

class ObjectDoesNotExist(Exception):
    pass

class MultipleObjectsReturned(Exception):
    pass



from typing import Any

def new(
    nbits,
    prefix: Any = ...,
    suffix: Any = ...,
    initial_value: int = ...,
    overflow: int = ...,
    little_endian: bool = ...,
    allow_wraparound: bool = ...,
    disable_shortcut: bool = ...,
): ...



# pyre-unsafe

from typing import Any, Mapping, Optional, Tuple, Union

from django.urls import URLPattern, URLResolver

def include(arg: Any, namespace: Optional[str] = ...) -> Tuple[Any, str, str]: ...
def path(
    route: str,
    view: Any,
    kwargs: Optional[Mapping[str, Any]] = ...,
    name: Optional[str] = ...,
) -> Union[URLPattern, URLResolver]: ...



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import copy
import io
import logging
import os
import re
import sys
import threading
import time
from types import TracebackType
from typing import Iterable, Optional, Pattern, Sequence


PERFORMANCE: int = 15
PROMPT: int = 50
SUCCESS: int = 60


LOG: logging.Logger = logging.getLogger(__name__)


stdout: io.StringIO = io.StringIO(newline="")
__handler: Optional["TimedStreamHandler"] = None


class Color:
    YELLOW: str = "\033[33m"
    RED: str = "\033[31m"
    GREEN: str = "\033[32m"


class Format:
    BOLD: str = "\033[1m"

    CLEAR_LINE: str = "\x1b[0G\x1b[K"
    CLEAR: str = "\033[0m"
    TRUNCATE_OVERFLOW: str = "\033[?7l"
    WRAP_OVERFLOW: str = "\033[?7h"
    NEWLINE: str = "\n"

    CURSOR_UP_LINE: str = "\x1b[1A"
    HIDE_CURSOR: str = "\x1b[?25l"
    SHOW_CURSOR: str = "\x1b[?25h"


class Character:
    LAMBDA: str = "ƛ"


class SectionFormatter(logging.Formatter):
    def __init__(self) -> None:
        super(SectionFormatter, self).__init__("%(asctime)s %(levelname)s %(message)s")

    def format(self, record: logging.LogRecord) -> str:
        formatted = super(SectionFormatter, self).format(record)
        return re.sub(r"DEBUG \[(.*)\]", r"\1", formatted)


class TimedStreamHandler(logging.StreamHandler):
    THRESHOLD: float = 0.5
    LINE_BREAKING_LEVELS: Sequence[str] = ["ERROR", "WARNING", "SUCCESS"]

    _terminate: bool = False
    _last_update: float = 0.0

    def __init__(self) -> None:
        super(TimedStreamHandler, self).__init__()
        self.setFormatter(logging.Formatter("%(message)s"))
        self.terminator: str = ""
        self.setLevel(logging.INFO)

        self._record: Optional[logging.LogRecord] = None
        self._active_lines: int = 0

        # Preamble preparing terminal.
        sys.stderr.write(
            Format.NEWLINE
            + Format.CLEAR_LINE
            + Format.CURSOR_UP_LINE
            + Format.HIDE_CURSOR
        )

        thread = threading.Thread(target=self._thread)
        thread.daemon = True
        thread.start()

    def clear_lines(self) -> str:
        if self._active_lines == 0:
            return ""
        return Format.CLEAR_LINE + "".join(
            [
                Format.CURSOR_UP_LINE + Format.CLEAR_LINE
                for n in range(self._active_lines - 1)
            ]
        )

    def emit(self, record: logging.LogRecord, age: Optional[float] = None) -> None:
        suffix = ""
        color = ""
        active_lines = record.msg.count("\n") + 1
        truncate = Format.TRUNCATE_OVERFLOW
        if record.levelname in self.LINE_BREAKING_LEVELS:
            record.msg += "\n"

        if record.levelname == "ERROR":
            color = Color.RED
            self._record = None
            active_lines = 0
            truncate = Format.WRAP_OVERFLOW
        elif record.levelname == "WARNING":
            color = Color.YELLOW
            self._record = None
            active_lines = 0
            truncate = Format.WRAP_OVERFLOW
        elif record.levelname == "PROMPT":
            color = Color.YELLOW
            self._record = None
            active_lines = 0
            truncate = Format.WRAP_OVERFLOW
        elif record.levelname == "SUCCESS":
            self._record = None
            active_lines = 0
            truncate = Format.WRAP_OVERFLOW
        elif age:
            if age > 10:
                color = Color.YELLOW
            if age > 30:
                color = Color.RED
            suffix = " {}[{:.1f}s]{}".format(
                color if color else "", age, Format.CLEAR if color else ""
            )
        else:
            self._record = record
            self._last_update = time.time()

        timed_record = copy.copy(record)
        timed_record.msg = (
            "{clear_line}{color} {cursor}{clear} {truncate}{message}{suffix}"
        ).format(
            clear_line=self.clear_lines(),
            color=color,
            cursor=Character.LAMBDA,
            clear=Format.CLEAR,
            truncate=truncate,
            message=record.msg,
            suffix=suffix,
        )
        self._active_lines = active_lines
        super(TimedStreamHandler, self).emit(timed_record)

    def _thread(self) -> None:
        while not self._terminate:
            record = self._record
            if record:
                age = time.time() - self._last_update
                if age > self.THRESHOLD:
                    self.emit(record, age)
            time.sleep(0.1)

    def terminate(self) -> None:
        self._terminate = True

        if self._active_lines > 0:
            sys.stderr.write(self.clear_lines())
            self._active_lines = 0

        # Reset terminal.
        sys.stderr.write(Format.WRAP_OVERFLOW + Format.SHOW_CURSOR)
        sys.stderr.flush()


def initialize(noninteractive: bool) -> None:
    global __handler

    if __handler:
        LOG.debug("Log handler already exists, skipping initialization.")
        return
    if noninteractive:
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(SectionFormatter())
        stream_handler.setLevel(logging.DEBUG)
        __handler = None
    else:
        stream_handler = TimedStreamHandler()
        __handler = stream_handler

    logging.addLevelName(PERFORMANCE, "PERFORMANCE")
    logging.addLevelName(PROMPT, "PROMPT")
    logging.addLevelName(SUCCESS, "SUCCESS")

    logging.basicConfig(level=logging.DEBUG, handlers=[stream_handler])


def start_logging_to_directory(noninteractive: bool, log_directory: str) -> None:
    if not noninteractive and log_directory is not None:
        if not os.path.exists(log_directory):
            os.makedirs(log_directory)
        handler = logging.FileHandler(os.path.join(log_directory, "pyre.stderr"))
        handler.setFormatter(SectionFormatter())
        handler.setLevel(logging.DEBUG)
        logger = logging.getLogger()
        logger.addHandler(handler)


def cleanup() -> None:
    global __handler
    handler = __handler
    if handler:
        handler.terminate()
        __handler = None

    output = stdout.getvalue()
    if output:
        sys.stdout.write(output)
        if not output.endswith("\n"):
            sys.stdout.write("\n")


class StreamLogger:
    _should_stop_reading_stream = False

    _server_log_pattern: Pattern[str] = re.compile(
        r"\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2} (\w+) (.*)"
    )

    def __init__(self, stream: Iterable[str]) -> None:
        self._reader = threading.Thread(target=self._read_stream, args=(stream,))
        self._reader.daemon = True

    def join(self) -> None:
        self._reader.join()

    def _log_server_stderr_message(self, server_message: str) -> None:
        line = server_message.rstrip()
        match = self._server_log_pattern.match(line)
        if match:
            section = match.groups()[0]
            message = match.groups()[1]
            if section == "ERROR":
                LOG.error(message)
            elif section == "INFO":
                LOG.info(message)
            elif section == "DUMP":
                LOG.warning(message)
            elif section == "WARNING":
                LOG.warning(message)
            elif section == "PROGRESS":
                LOG.info(message)
            elif section == "PARSER":
                LOG.error(message)
            else:
                LOG.debug("[%s] %s", section, message)
        else:
            LOG.debug(line)

    def _read_stream(self, stream: Iterable[str]) -> None:
        try:
            for line in stream:
                if self._should_stop_reading_stream:
                    return
                self._log_server_stderr_message(line)
        except Exception:
            pass

    def __enter__(self) -> "StreamLogger":
        self._should_stop_reading_stream = False
        self._reader.start()
        return self

    def __exit__(
        self,
        _type: Optional[BaseException],
        _value: Optional[BaseException],
        _traceback: Optional[TracebackType],
    ) -> None:
        self._should_stop_reading_stream = True


def get_yes_no_input(prompt: str) -> bool:
    choice = get_input(prompt, suffix=" [Y/n] ")
    return choice.lower() in ["", "y", "ye", "yes"]


def get_optional_input(prompt: str, default: str) -> str:
    result = get_input(prompt, suffix=" (Default: `{}`): ".format(default))
    if result == "":
        return default
    return result


def get_input(prompt: str, suffix: str = "") -> str:
    LOG.log(PROMPT, prompt + suffix)
    return input().strip()



import sys
import xml.sax
from typing import IO, Any, Iterable, List, NoReturn, Optional, Text, Union
from xml.sax.handler import ContentHandler, ErrorHandler
from xml.sax.xmlreader import InputSource, Locator

class SAXException(Exception):
    def __init__(self, msg: str, exception: Optional[Exception] = ...) -> None: ...
    def getMessage(self) -> str: ...
    def getException(self) -> Exception: ...
    def __getitem__(self, ix: Any) -> NoReturn: ...

class SAXParseException(SAXException):
    def __init__(self, msg: str, exception: Exception, locator: Locator) -> None: ...
    def getColumnNumber(self) -> int: ...
    def getLineNumber(self) -> int: ...
    def getPublicId(self): ...
    def getSystemId(self): ...

class SAXNotRecognizedException(SAXException): ...
class SAXNotSupportedException(SAXException): ...
class SAXReaderNotAvailable(SAXNotSupportedException): ...

default_parser_list: List[str]

if sys.version_info >= (3, 8):
    def make_parser(parser_list: Iterable[str] = ...) -> xml.sax.xmlreader.XMLReader: ...

else:
    def make_parser(parser_list: List[str] = ...) -> xml.sax.xmlreader.XMLReader: ...

def parse(
    source: Union[str, IO[str], IO[bytes]],
    handler: xml.sax.handler.ContentHandler,
    errorHandler: xml.sax.handler.ErrorHandler = ...,
) -> None: ...
def parseString(
    string: Union[bytes, Text],
    handler: xml.sax.handler.ContentHandler,
    errorHandler: Optional[xml.sax.handler.ErrorHandler] = ...,
) -> None: ...
def _create_parser(parser_name: str) -> xml.sax.xmlreader.XMLReader: ...



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from collections import deque
from typing import Tuple

from .models import DBID, SharedText, SharedTextKind
from .pipeline import PipelineStep, Summary
from .trace_graph import TraceGraph


log = logging.getLogger("sapp")


class AddIssueInstanceLeaves(PipelineStep[TraceGraph, TraceGraph]):
    """For all issues with a given code, adds a leaf at the issue frame and adds
    the relevant trace_frame_leaf_assocs. The depth increases as the trace
    frames get further away (in the caller->callee direction) from the issue
    frame. Useful for queries in the callee->caller direction."""

    def __init__(self, code: int, leaf_name: str, leaf_kind: SharedTextKind) -> None:
        super().__init__()
        self.code = code
        self.leaf_name = leaf_name
        self.leaf_kind = leaf_kind

    def run(self, input: TraceGraph, summary: Summary) -> Tuple[TraceGraph, Summary]:
        graph = input

        # Get all the issue instances within this category
        instances = [
            instance
            for instance in graph.get_issue_instances()
            if self.code == graph.get_issue(instance.issue_id).code
        ]

        # Add all the trace frame ids for all the issue instances
        trace_frames = []
        for instance in instances:
            trace_frames.extend(graph.get_issue_instance_trace_frames(instance))

        # Explore forward (caller -> callee; issue -> leaf)
        queue = deque([(frame, 0) for frame in trace_frames])
        depth_by_frame_id = {}
        while len(queue) > 0:
            trace_frame, depth = queue.popleft()
            trace_frame_id = trace_frame.id.local_id

            # Skip repeat frames unless we arrived at them by a shorter path.
            if (
                trace_frame_id in depth_by_frame_id
                and depth >= depth_by_frame_id[trace_frame_id]
            ):
                continue
            else:
                # Record the minimum depth.
                depth_by_frame_id[trace_frame_id] = depth

            # Iterate forward
            queue.extend(
                (next_frame, depth + 1)
                for next_frame in graph.get_next_trace_frames(trace_frame)
            )

        # Create new leaves based on these depths
        leaf = graph.get_shared_text(self.leaf_kind, self.leaf_name)
        if leaf is None:
            leaf = SharedText.Record(
                id=DBID(), contents=self.leaf_name, kind=self.leaf_kind
            )
            graph.add_shared_text(leaf)

        # Add the assoc to the leaf
        log.info(
            'Adding %d "%s" leaves from issues with code %d...',
            len(depth_by_frame_id),
            self.leaf_name,
            self.code,
        )
        for trace_frame_id, depth in depth_by_frame_id.items():
            trace_frame = graph.get_trace_frame_from_id(trace_frame_id)
            graph.add_trace_frame_leaf_assoc(trace_frame, leaf, depth)

        return graph, summary



# pyre-unsafe

from re import Pattern
from types import ModuleType
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import re2
from django.http import Http404, HttpRequest, HttpResponse

class Resolver404(Http404): ...

class ResolverMatch:
    kwargs: Dict[str, Any] = ...
    view_name: str = ...
    func: Any = ...
    args: Any = ...
    url_name: str = ...
    app_name: str = ...
    _func_path: str = ...
    view_path: str = ...
    def __getitem__(self, index: int) -> Any: ...

class NoReverseMatch(Exception): ...

def reverse(
    viewname: Any,
    urlconf: Any = ...,
    args: Optional[Sequence[Any]] = ...,
    kwargs: Optional[Mapping[str, Any]] = ...,
    prefix: Any = ...,
    current_app: Any = ...,
) -> str: ...
def resolve(path: str, urlconf: Any = ...) -> ResolverMatch: ...

class RegexURLPattern:
    regex: Pattern
    callback: Callable[..., Any]
    name: Optional[str]

class RegexURLResolver:
    url_patterns: Iterable[Union[RegexURLPattern, "RegexURLResolver"]]
    urlconf_module: Any
    urlconf_name: Union[str, ModuleType]
    regex: Pattern
    def resolve(self, path: str) -> ResolverMatch: ...
    def resolve_error_handler(
        self, status_code: int
    ) -> Tuple[Callable[[HttpRequest, ...], HttpResponse], Dict]: ...
    # this is not in Django but we add it on
    _re_set: Union[bool, Optional[re2.Set]]

def get_urlconf(default: Any = None) -> Any: ...
def get_resolver(urlconf: Any) -> RegexURLResolver: ...
def get_callable(view_name: str) -> Callable[..., Awaitable[HttpResponse]]: ...



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import logging
from logging import Logger
from typing import List, Optional

from .. import json_rpc
from ..analysis_directory import AnalysisDirectory
from ..configuration import Configuration
from ..project_files_monitor import ProjectFilesMonitor
from .command import (
    CommandArguments,
    ExitCode,
    IncrementalStyle,
    Result,
    State,
    typeshed_search_path,
)
from .reporting import Reporting
from .start import Start


LOG: Logger = logging.getLogger(__name__)


class Incremental(Reporting):
    NAME = "incremental"

    def __init__(
        self,
        command_arguments: CommandArguments,
        original_directory: str,
        *,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
        nonblocking: bool,
        incremental_style: IncrementalStyle,
        no_start_server: bool,
        no_watchman: bool,
    ) -> None:
        super(Incremental, self).__init__(
            command_arguments, original_directory, configuration, analysis_directory
        )
        self._nonblocking = nonblocking
        self._incremental_style = incremental_style
        self._no_start_server = no_start_server
        self._no_watchman = no_watchman

    @staticmethod
    def from_arguments(
        arguments: argparse.Namespace,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> "Incremental":
        return Incremental(
            CommandArguments.from_arguments(arguments),
            original_directory,
            configuration=configuration,
            analysis_directory=analysis_directory,
            nonblocking=arguments.nonblocking,
            incremental_style=arguments.incremental_style,
            no_start_server=arguments.no_start,
            no_watchman=getattr(arguments, "no_watchman", False),
        )

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        incremental_help = """
        Connects to a running Pyre server and returns the current type errors for your
        project. If no server exists for your projects, starts a new one. Running `pyre`
        implicitly runs `pyre incremental`.

        By default, incremental checks ensure that all dependencies of changed files are
        analyzed before returning results. If you'd like to get partial type checking
        results eagerly, you can run `pyre incremental --nonblocking`.
        """
        incremental = parser.add_parser(cls.NAME, epilog=incremental_help)
        incremental.set_defaults(command=cls.from_arguments)
        incremental.add_argument(
            "--nonblocking",
            action="store_true",
            help=(
                "Ask the server to return partial results immediately, "
                "even if analysis is still in progress."
            ),
        )
        incremental.add_argument(
            "--incremental-style",
            type=IncrementalStyle,
            choices=list(IncrementalStyle),
            default=IncrementalStyle.FINE_GRAINED,
            help="How to approach doing incremental checks.",
        )
        incremental.add_argument(
            "--no-start", action="store_true", help=argparse.SUPPRESS
        )
        # This is mostly to allow `restart` to pass on the flag to `start`.
        incremental.add_argument(
            "--no-watchman", action="store_true", help=argparse.SUPPRESS
        )

    def _run(self) -> None:
        if (not self._no_start_server) and self._state() == State.DEAD:
            LOG.info("Starting server at `%s`.", self._analysis_directory.get_root())
            exit_code = (
                Start(
                    self._command_arguments,
                    self._original_directory,
                    configuration=self._configuration,
                    analysis_directory=self._analysis_directory,
                    terminal=False,
                    store_type_check_resolution=False,
                    use_watchman=not self._no_watchman,
                    incremental_style=self._incremental_style,
                )
                .run()
                .exit_code()
            )
            if exit_code != ExitCode.SUCCESS:
                self._exit_code = ExitCode.FAILURE
                return
        else:
            self._restart_file_monitor_if_needed()

        if self._state() != State.DEAD:
            LOG.info("Waiting for server...")

        with self._analysis_directory.acquire_shared_reader_lock():
            request = json_rpc.Request(
                method="displayTypeErrors", parameters={"files": []}
            )
            self._send_and_handle_socket_request(request, self._version_hash)

    def _socket_result_handler(self, result: Result) -> None:
        errors = self._get_errors(result)
        self._print(errors)
        if errors:
            self._exit_code = ExitCode.FOUND_ERRORS

    def _flags(self) -> List[str]:
        flags = super()._flags()
        flags.extend(["-expected-binary-version", self._configuration.version_hash])

        search_path = self._configuration.search_path + typeshed_search_path(
            self._configuration.typeshed
        )
        if search_path:
            flags.extend(["-search-path", ",".join(search_path)])

        excludes = self._configuration.excludes
        for exclude in excludes:
            flags.extend(["-exclude", exclude])

        extensions = self._configuration.extensions
        for extension in extensions:
            flags.extend(["-extension", extension])

        if self._nonblocking:
            flags.append("-nonblocking")

        return flags

    def _restart_file_monitor_if_needed(self) -> None:
        if self._no_watchman:
            return
        ProjectFilesMonitor.restart_if_dead(
            self._configuration, self._project_root, self._analysis_directory
        )



# pyre-unsafe

from typing import Any, TypeVar

_T = TypeVar("_T")

def curry(_curried_func, *args, **kwargs) -> Any: ...
def memoize(func: _T, cache, num_args) -> _T: ...

class cached_property:
    func: Any
    name: str
    def __get__(self, instance, type=...) -> Any: ...

# In the future, we'll properly type promises as parameteric types.
Promise = Any

def lazy(func: _T, *resultclasses) -> _T: ...

class LazyObject: ...

def partition(predicate: Any, values: Any) -> Any: ...



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

#!/usr/bin/env python3

import logging
from typing import Optional, Tuple

from .bulk_saver import BulkSaver
from .db import DB
from .decorators import log_time
from .models import (
    Issue,
    IssueInstanceSharedTextAssoc,
    MetaRunToRunAssoc,
    PrimaryKeyGenerator,
    Run,
    RunStatus,
    RunSummary,
    TraceFrame,
    TraceFrameAnnotation,
    TraceKind,
)
from .pipeline import PipelineStep, Summary
from .trace_graph import TraceGraph


log = logging.getLogger("sapp")


# pyre-fixme[13]: Attribute `summary` is never initialized.
class DatabaseSaver(PipelineStep[TraceGraph, RunSummary]):
    RUN_MODEL = Run

    def __init__(
        self,
        database: DB,
        use_lock: bool = False,
        primary_key_generator: Optional[PrimaryKeyGenerator] = None,
    ):
        self.use_lock = use_lock
        self.dbname = database.dbname
        self.database = database
        self.primary_key_generator = primary_key_generator or PrimaryKeyGenerator()
        self.bulk_saver = BulkSaver(self.primary_key_generator)
        self.summary: Summary

    @log_time
    def run(self, input: TraceGraph, summary: Summary) -> Tuple[RunSummary, Summary]:
        # pyre-fixme[16]: `DatabaseSaver` has no attribute `graph`.
        self.graph = input
        self.summary = summary

        self._prep_save()
        return self._save(), self.summary

    def _prep_save(self):
        """ Prepares the bulk saver to load the trace graph info into the
        database.
        """
        log.info("Preparing bulk save.")
        self.graph.update_bulk_saver(self.bulk_saver)

        for trace_kind, unused in self.summary["trace_entries"].items():
            log.info(
                "Dropped %d unused %s, %d are missing",
                sum(len(v) for v in unused.values()),
                trace_kind,
                len(self.summary["missing_traces"][trace_kind]),
            )

    def _save(self) -> RunSummary:
        """ Saves bulk saver's info into the databases in bulk.
        """
        assert self.summary["run"] is not None, "Must have called process before"

        trace_frames = self.bulk_saver.get_items_to_add(TraceFrame)
        log.info(
            "Saving %d issues, %d trace frames, %d trace annotations",
            len(self.bulk_saver.get_items_to_add(Issue)),
            len(self.bulk_saver.get_items_to_add(TraceFrame)),
            len(self.bulk_saver.get_items_to_add(TraceFrameAnnotation)),
        )

        num_pre = 0
        num_post = 0
        for frame in trace_frames:
            if frame.kind == TraceKind.PRECONDITION:
                num_pre += 1
            elif frame.kind == TraceKind.POSTCONDITION:
                num_post += 1
        log.info(
            "Within trace frames: %d preconditions, %d postconditions",
            num_pre,
            num_post,
        )

        with self.database.make_session() as session:
            pk_gen = self.primary_key_generator.reserve(
                session, [Run], use_lock=self.use_lock
            )
            self.summary["run"].id.resolve(id=pk_gen.get(Run), is_new=True)
            session.add(self.summary["run"])
            meta_run_identifier = self.summary.get("meta_run_identifier")
            if meta_run_identifier is not None:
                session.add(
                    # pyre-ignore[28]: SQLAlchemy and Pysa don't get along.
                    MetaRunToRunAssoc(
                        meta_run_id=meta_run_identifier, run_id=self.summary["run"].id
                    )
                )
            session.commit()

            run_id = self.summary["run"].id.resolved()
            self.summary["run"] = None  # Invalidate it

        self.bulk_saver.save_all(self.database, self.use_lock)

        # Now that the run is finished, fetch it from the DB again and set its
        # status to FINISHED.
        with self.database.make_session() as session:
            run = session.query(self.RUN_MODEL).filter_by(id=run_id).one()
            run.status = RunStatus.FINISHED
            session.add(run)
            session.commit()
            run_summary = run.get_summary()

        run_summary.num_invisible_issues = 0
        run_summary.num_missing_preconditions = len(
            self.summary["missing_traces"][TraceKind.precondition]
        )
        run_summary.num_missing_postconditions = len(
            self.summary["missing_traces"][TraceKind.postcondition]
        )

        return run_summary



# pyre-unsafe

import datetime
from typing import Dict

from django.db.backends.utils import CursorWrapper

class BaseDatabaseWrapper:
    alias: str
    settings_dict: Dict[str, object]
    def __init__(self, settings_dict, alias=...): ...
    def cursor(self) -> CursorWrapper: ...
    def close(self): ...
    def close_if_unusable_or_obsolete(self): ...
    def connection(self) -> ...: ...



# pyre-unsafe

from typing import Any, Callable

from django.core.urlresolvers import RegexURLPattern, RegexURLResolver

def include(
    arg: Any, namespace: str = ..., app_name: str = ...
) -> RegexURLResolver: ...
def url(
    regex: str, view: Callable, kwargs: Any = ..., name: str = ..., prefix: str = ...
) -> RegexURLPattern: ...



# Stubs for email.charset (Python 3.4)

from typing import Any, Iterator, List, Optional

QP: int  # undocumented
BASE64: int  # undocumented
SHORTEST: int  # undocumented

class Charset:
    input_charset: str
    header_encoding: int
    body_encoding: int
    output_charset: Optional[str]
    input_codec: Optional[str]
    output_codec: Optional[str]
    def __init__(self, input_charset: str = ...) -> None: ...
    def get_body_encoding(self) -> str: ...
    def get_output_charset(self) -> Optional[str]: ...
    def header_encode(self, string: str) -> str: ...
    def header_encode_lines(self, string: str, maxlengths: Iterator[int]) -> List[str]: ...
    def body_encode(self, string: str) -> str: ...
    def __str__(self) -> str: ...
    def __eq__(self, other: Any) -> bool: ...
    def __ne__(self, other: Any) -> bool: ...

def add_charset(
    charset: str, header_enc: Optional[int] = ..., body_enc: Optional[int] = ..., output_charset: Optional[str] = ...
) -> None: ...
def add_alias(alias: str, canonical: str) -> None: ...
def add_codec(charset: str, codecname: str) -> None: ...



# pyre-unsafe

from unittest import TestCase

from django.test.client import Client
from django.test.utils import override_settings

class SimpleTestCase(TestCase):
    client: testing.client.base.TestClient = ...
    @contextlib.contextmanager
    def settings(self, **kwargs) -> None: ...
